/*
 * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 *
 * (c) Copyright 1996 - 2001 Gary Henderson (gary@daniver.demon.co.uk) and
 *                           Jerremy Koot (jkoot@snes9x.com)
 *
 * Super FX C emulator code 
 * (c) Copyright 1997 - 1999 Ivar (Ivar@snes9x.com) and
 *                           Gary Henderson.
 * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
 *
 * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
 * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
 * C4 C code (c) Copyright 2001 Gary Henderson (gary@daniver.demon.co.uk).
 *
 * DOS port code contains the works of other authors. See headers in
 * individual files.
 *
 * Snes9x homepage: www.snes9x.com
 *
 * Permission to use, copy, modify and distribute Snes9x in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Snes9x is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Snes9x or software derived from Snes9x.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so everyone can benefit from the modifications
 * in future versions.
 *
 * Super NES and Super Nintendo Entertainment System are trademarks of
 * Nintendo Co., Limited and its subsidiary companies.
 */
#ifdef __DJGPP
#include <allegro.h>
#undef TRUE
#endif

#include "snes9x.h"
#include "spc700.h"
#include "apu.h"
#include "soundux.h"
#include "snesapu.h"
#include "cpuexec.h"

/* For note-triggered SPC dump support */
#include "snapshot.h"

#ifdef __UOSNES__
extern const char *S9xGetSPCFilename ();
#else
extern "C" {const char *S9xGetFilenameInc (const char *);}
#endif

int spc_is_dumping=0;
int spc_is_dumping_temp;
uint8 spc_dump_dsp[0x100]; 

extern int NoiseFreq [32];
#ifdef DEBUGGER
void S9xTraceSoundDSP (const char *s, int i1 = 0, int i2 = 0, int i3 = 0,
		       int i4 = 0, int i5 = 0, int i6 = 0, int i7 = 0);
#endif

bool8 S9xInitAPU ()
{
    IAPU.RAM = (uint8 *) malloc (0x10000);
    IAPU.ShadowRAM = (uint8 *) malloc (0x10000);
    IAPU.CachedSamples = (uint8 *) malloc (0x40000);
    
    if (!IAPU.RAM || !IAPU.ShadowRAM || !IAPU.CachedSamples)
    {
		S9xDeinitAPU ();
		return (FALSE);
    }

	memset(IAPU.RAM, 0, 0x10000);
	memset(IAPU.ShadowRAM, 0, 0x10000);
	memset(IAPU.CachedSamples, 0, 0x40000);

	//SNESAPU
    InitAPUDSP();

    return (TRUE);
}

void S9xDeinitAPU ()
{
    if (IAPU.RAM)
    {
		free ((char *) IAPU.RAM);
		IAPU.RAM = NULL;
    }
    if (IAPU.ShadowRAM)
    {
		free ((char *) IAPU.ShadowRAM);
		IAPU.ShadowRAM = NULL;
    }
    if (IAPU.CachedSamples)
    {
		free ((char *) IAPU.CachedSamples);
		IAPU.CachedSamples = NULL;
    }
}

EXTERN_C uint8 APUROM [64];

void S9xResetAPU ()
{
    int i;

    Settings.APUEnabled = Settings.NextAPUEnabled;
    if(SNESGameFixes.APU_Disable) Settings.APUEnabled = FALSE;

	ZeroMemory(spc_dump_dsp, 0x100);
	//ZeroMemory(IAPU.RAM, 0x100);
    memset (IAPU.RAM, Options.APURAMInitialValue, 0x100);
	memset(IAPU.RAM+0x20, 0xFF, 0x20);
	memset(IAPU.RAM+0x60, 0xFF, 0x20);
	memset(IAPU.RAM+0xA0, 0xFF, 0x20);
	memset(IAPU.RAM+0xE0, 0xFF, 0x20);

	for(i=1;i<256;i++)
	{
		memcpy(IAPU.RAM+(i<<8), IAPU.RAM, 0x100);
	}

    memcpy (IAPU.ShadowRAM, IAPU.RAM, 0x10000);
	
    ZeroMemory (IAPU.CachedSamples, 0x40000);
    ZeroMemory (APU.OutPorts, 4);
    IAPU.DirectPage = IAPU.RAM;
    memmove (&IAPU.RAM [0xffc0], APUROM, sizeof (APUROM));
    memmove (APU.ExtraRAM, APUROM, sizeof (APUROM));
    IAPU.PC = IAPU.RAM + IAPU.RAM [0xfffe] + (IAPU.RAM [0xffff] << 8);
    APU.Cycles = 0;
    APURegisters.YA.W = 0;
    APURegisters.X = 0;
    APURegisters.S = 0xff;
    APURegisters.P = 0;
    S9xAPUUnpackStatus ();
    APURegisters.PC = 0;
    IAPU.APUExecuting = Settings.APUEnabled;
#ifdef SPC700_SHUTDOWN
    IAPU.WaitAddress1 = NULL;
    IAPU.WaitAddress2 = NULL;
    IAPU.WaitCounter = 1;
#endif

	S9xSetAPUTimerCounter();
    IAPU.WhichEvent = 0;
    IAPU.NextAPUTimerPos = 0;
	IAPU.APUTimerCounter_err = 0;

	int32 t = IAPU.APUTimerCounter + IAPU.APUTimerCounter_err;
    IAPU.NextAPUTimerPos += (t >> 16);
	IAPU.APUTimerCounter_err = (t & 0xffff);

    APU.ShowROM = TRUE;
    IAPU.RAM [0xf1] = 0x80;
		
    for (i = 0; i < 3; i++)
    {
		APU.TimerEnabled [i] = FALSE;
		APU.TimerValueWritten [i] = 0;
		APU.TimerTarget [i] = 0;
		APU.Timer [i] = 0;
    }
    for (int j = 0; j < 0x80; j++)
		APU.DSP [j] = 0;
	
    IAPU.TwoCycles = IAPU.OneCycle * 2;
	
    for (i = 0; i < 256; i++)
		S9xAPUCycles [i] = S9xAPUCycleLengths [i] * IAPU.OneCycle;
	
    APU.DSP [APU_ENDX] = 0;
    APU.DSP [APU_KOFF] = 0;
    APU.DSP [APU_KON] = 0;
    //APU.DSP [APU_FLG] = APU_MUTE | APU_ECHO_DISABLED;
	APU.DSP [APU_FLG] = APU_SOFT_RESET | APU_MUTE;
    APU.KeyedChannels = 0;
	
    S9xResetSound (TRUE);
    S9xSetEchoEnable (0);

	//SNESAPU
	ResetAPUDSP();
}

void S9xSetAPUDSP (uint8 byte)
{
	APUDSPIn(IAPU.RAM [0xf2], byte);
}

void S9xSetAPUControl (uint8 byte)
{
	//if (byte & 0x40)
	//printf ("*** Special SPC700 timing enabled\n");
    if ((byte & 1) != 0 && !APU.TimerEnabled [0])
    {
		APU.Timer [0] = 0;
		IAPU.RAM [0xfd] = 0;
		if ((APU.TimerTarget [0] = IAPU.RAM [0xfa]) == 0)
			APU.TimerTarget [0] = 0x100;
    }
    if ((byte & 2) != 0 && !APU.TimerEnabled [1])
    {
		APU.Timer [1] = 0;
		IAPU.RAM [0xfe] = 0;
		if ((APU.TimerTarget [1] = IAPU.RAM [0xfb]) == 0)
			APU.TimerTarget [1] = 0x100;
    }
    if ((byte & 4) != 0 && !APU.TimerEnabled [2])
    {
		APU.Timer [2] = 0;
		IAPU.RAM [0xff] = 0;
		if ((APU.TimerTarget [2] = IAPU.RAM [0xfc]) == 0)
			APU.TimerTarget [2] = 0x100;
    }
    APU.TimerEnabled [0] = byte & 1;
    APU.TimerEnabled [1] = (byte & 2) >> 1;
    APU.TimerEnabled [2] = (byte & 4) >> 2;
	
    if (byte & 0x10)
		IAPU.RAM [0xF4] = IAPU.RAM [0xF5] = 0;
	
    if (byte & 0x20)
		IAPU.RAM [0xF6] = IAPU.RAM [0xF7] = 0;
	
    if (byte & 0x80)
    {
		if (!APU.ShowROM)
		{
			memmove (&IAPU.RAM [0xffc0], APUROM, sizeof (APUROM));
			APU.ShowROM = TRUE;
		}
    }
    else
    {
		if (APU.ShowROM)
		{
			APU.ShowROM = FALSE;
			memmove (&IAPU.RAM [0xffc0], APU.ExtraRAM, sizeof (APUROM));
		}
    }
    IAPU.RAM [0xf1] = byte;
}

void S9xSetAPUTimer (uint16 Address, uint8 byte)
{
    IAPU.RAM [Address] = byte;
	
    switch (Address)
    {
    case 0xfa:
		if ((APU.TimerTarget [0] = IAPU.RAM [0xfa]) == 0)
			APU.TimerTarget [0] = 0x100;
		APU.TimerValueWritten [0] = TRUE;
		break;
    case 0xfb:
		if ((APU.TimerTarget [1] = IAPU.RAM [0xfb]) == 0)
			APU.TimerTarget [1] = 0x100;
		APU.TimerValueWritten [1] = TRUE;
		break;
    case 0xfc:
		if ((APU.TimerTarget [2] = IAPU.RAM [0xfc]) == 0)
			APU.TimerTarget [2] = 0x100;
		APU.TimerValueWritten [2] = TRUE;
		break;
    }
}

void S9xSetAPUTimerCounter(void)
{
	IAPU.APUTimerCounter =
		(int32)(((uint64)SNES_CYCLES_PER_SECOND * 0x10000 * Settings.H_Max_normal)
				/ (SNES_CYCLES_PER_SCANLINE * 64000));
	return;
}

static inline void APUTimerCheck()
{
	if (APU.TimerEnabled [2]) {
		APU.Timer [2]++;
		if (APU.Timer [2] >= APU.TimerTarget [2]) {
			IAPU.RAM [0xff] = (IAPU.RAM [0xff] + 1) & 0xf;
			APU.Timer [2] -= APU.TimerTarget [2];
#ifdef SPC700_SHUTDOWN
			IAPU.WaitCounter = 1;
			IAPU.APUExecuting = Settings.APUEnabled;
#endif
		}
	}
	if (IAPU.WhichEvent == 7) {
		if (APU.TimerEnabled [0]) {
			APU.Timer [0]++;
			if (APU.Timer [0] >= APU.TimerTarget [0]) {
				IAPU.RAM [0xfd] = (IAPU.RAM [0xfd] + 1) & 0xf;
				APU.Timer [0] -= APU.TimerTarget [0];
#ifdef SPC700_SHUTDOWN
				IAPU.WaitCounter = 1;
				IAPU.APUExecuting = Settings.APUEnabled;
#endif
			}
		}
		if (APU.TimerEnabled [1]) {
			APU.Timer [1]++;
			if (APU.Timer [1] >= APU.TimerTarget [1]) {
				IAPU.RAM [0xfe] = (IAPU.RAM [0xfe] + 1) & 0xf;
				APU.Timer [1] -= APU.TimerTarget [1];
#ifdef SPC700_SHUTDOWN
				IAPU.WaitCounter = 1;
				IAPU.APUExecuting = Settings.APUEnabled;
#endif
			}
		}
	}
}

void S9xUpdateAPUTimer(void)
{
	while(CPU.Cycles >= IAPU.NextAPUTimerPos) {
		if(Settings.SoundSync && (IAPU.WhichEvent & 3) == 3)
			S9xGenerateSound ();

		APUTimerCheck();

		IAPU.WhichEvent = ((IAPU.WhichEvent + 1) & 7);
		int32 t = IAPU.APUTimerCounter + IAPU.APUTimerCounter_err;
		IAPU.NextAPUTimerPos += (t >> 16);
		IAPU.APUTimerCounter_err = (t & 0xffff);
		
		//SNESAPU
		t64Cnt++;
	}
}

void S9xUpdateAPUTimerCPU(void)
{
	while(CPU.Cycles >= IAPU.NextAPUTimerPos) {
		if (IAPU.APUExecuting) {
			while (APU.Cycles < IAPU.NextAPUTimerPos)
				APU_EXECUTE1();
		}
		else {
			APU.Cycles = IAPU.NextAPUTimerPos;
		}

		if(Settings.SoundSync && (IAPU.WhichEvent & 3) == 3)
			S9xGenerateSound ();

		APUTimerCheck();

		IAPU.WhichEvent = ((IAPU.WhichEvent + 1) & 7);
		int32 t = IAPU.APUTimerCounter + IAPU.APUTimerCounter_err;
		IAPU.NextAPUTimerPos += (t >> 16);
		IAPU.APUTimerCounter_err = (t & 0xffff);
		
		//SNESAPU
		t64Cnt++;
	}
	APU_EXECUTE();
}

uint8 S9xGetAPUDSP ()
{
	// Writes to 80-FFh have no effect (reads are mirrored from lower mem)
	uint8 reg = IAPU.RAM [0xf2] & 0x7f;
	uint8 byte;

	switch (reg & 0x0f) {
	case APU_OUTX:
		{
			int i = reg >> 4;
			byte = (uint8)((mix[i].mFlg & (MFLG_OFF | MFLG_END)) ? 0 : (mix[i].mOut >> 8));
		}
		break;

	case APU_ENVX:
		{
			if(Settings.SoundEnvelopeHeightReading) {
				int i = reg >> 4;
				byte = (uint8)((mix[i].eVal >> E_SHIFT) & 0x7f);
			}
			else
				byte = 0;
		}
		break;

	default:
		byte = APU.DSP [reg];
		if(reg == APU_ENDX) {
			for(int i = 0; i < 8; i++) {
				if(mix[i].mFlg & MFLG_SSRC)
					byte &= ~(1 << i);
			}
		}
		break;
	}
	return byte;
}

