#include "snes9x.h"
#include "memmap.h"
#include "soundux.h"
#include "directx.h"
#include "uowinput.h"
#include "explorer.h"
#include <io.h>
#include <direct.h>
#include "cfgfile.h"
#include "cfgtpl.h"
#include "commutil.h"

extern char ROMFileName[];
extern char InitialROMFileDirectory[];
extern char RecentROMFileName[];
extern OPENFILENAMEEXPLORERPOS uosxplorerposition;

static bool CannotCreateIniFlag = 0;
static char cfgfilename[_MAX_PATH + 1];

//RevID for refresh config
static int RevID = 82;

static void write_config();

void read_config()
{
	//create fullpath of config.
	int len = strlen(__argv[0]);
	if(len > _MAX_PATH)
		return;

	strcpy(cfgfilename, __argv[0]);
	strcpy(cfgfilename + (len - 4), ".ini");
	
	if(!load_configfile(cfgfilename,
						config_data,
						write_config)) {
		if(!configfile_cannot_create) {
			write_config();
			flush_configfile();
		}
		return;
	}

    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fname [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];
	char tmpCharBuf[_MAX_PATH + 1];
	char *tmpcharptr;
	int i;

	/* Read [windows] */
	i = get_config_int((int)GUI.PriorityClass, Inis.windows, "PriorityClass");
	switch(i) {
	case 0: GUI.PriorityClass = IDLE_PRIORITY_CLASS; break;
	case 1: GUI.PriorityClass = NORMAL_PRIORITY_CLASS; break;
	case 2: GUI.PriorityClass = HIGH_PRIORITY_CLASS; break;
	case 3: GUI.PriorityClass = REALTIME_PRIORITY_CLASS; break;
	default: GUI.PriorityClass = NORMAL_PRIORITY_CLASS; break;
	}
	
	i = get_config_int((int)GUI.ThreadPriority, Inis.windows, "ThreadPriority");
	switch(i) {
	case 0: GUI.ThreadPriority = THREAD_PRIORITY_LOWEST; break;
	case 1: GUI.ThreadPriority = THREAD_PRIORITY_BELOW_NORMAL; break;
	case 2: GUI.ThreadPriority = THREAD_PRIORITY_NORMAL; break;
	case 3: GUI.ThreadPriority = THREAD_PRIORITY_ABOVE_NORMAL; break;
	case 4: GUI.ThreadPriority = THREAD_PRIORITY_HIGHEST; break;
	case 5: GUI.ThreadPriority = THREAD_PRIORITY_IDLE; break;
	case 6: GUI.ThreadPriority = THREAD_PRIORITY_TIME_CRITICAL; break;
	default: GUI.ThreadPriority = THREAD_PRIORITY_NORMAL; break;
	}

	GUI.WindowTopMost = (unsigned char)get_config_int((int)GUI.WindowTopMost, Inis.windows, "WindowTopMost");

	/* Read [graphics] */
	i = get_config_int((int)GUI.FullScreen, Inis.graphics, "Fullscreen");
	GUI.FullScreen = (i ? 1 : 0);
	GUI.DoubleRendering = (unsigned char)get_config_int((int)GUI.DoubleRendering, Inis.graphics, "DoubleRendering");
	i = get_config_int(GUI.FullscreenDisplayModeIndexNormal16, Inis.graphics, "FullscreenDisplayModeIndexNormal16");
	GUI.FullscreenDisplayModeIndexNormal16 = ((i >= 0 && i <= 27) ? i : GUI.FullscreenDisplayModeIndexNormal16);
	i = get_config_int(GUI.FullscreenDisplayModeIndexDouble16, Inis.graphics, "FullscreenDisplayModeIndexDouble16");
	GUI.FullscreenDisplayModeIndexDouble16 = ((i >= 12 && i <= 27) ? i : GUI.FullscreenDisplayModeIndexDouble16);
	i = get_config_int(GUI.FullscreenDisplayModeIndexNormal8, Inis.graphics, "FullscreenDisplayModeIndexNormal8");
	GUI.FullscreenDisplayModeIndexNormal8 = ((i >= 0 && i <= 27) ? i : GUI.FullscreenDisplayModeIndexNormal8);
	i = get_config_int(GUI.FullscreenDisplayModeIndexDouble8, Inis.graphics, "FullscreenDisplayModeIndexDouble8");
	GUI.FullscreenDisplayModeIndexDouble8 = ((i >= 12 && i <= 27) ? i : GUI.FullscreenDisplayModeIndexDouble8);
	i = get_config_int(GUI.DoubleRenderingModes16, Inis.graphics, "DoubleRenderingModes16");
	GUI.DoubleRenderingModes16 = ((i >= 1 && i <= 10) ? i : GUI.DoubleRenderingModes16);
	i = get_config_int(GUI.DoubleRenderingModes8, Inis.graphics, "DoubleRenderingModes8");
	GUI.DoubleRenderingModes8 = ((i >= 1 && i <= 3) ? i : GUI.DoubleRenderingModes8);
	DirectX.DisableHardwareScale = (unsigned char)get_config_int((int)DirectX.DisableHardwareScale, Inis.graphics, "DisableHardwareScale");
	GUI.dRendering = (unsigned char)get_config_int((int)GUI.dRendering, Inis.graphics, "DirectScreenRendering");
	i = get_config_int((int)GUI.DoubleBuffered, Inis.graphics, "DoubleBuffering");
	GUI.DoubleBuffered = i ? 1 : 0;
	GUI.ForceWindowedDirectDrawClipper = (unsigned char)get_config_int((int)GUI.ForceWindowedDirectDrawClipper,
																	   Inis.graphics,
																	   "ForceWindowedDirectDrawClipper");
	i = get_config_int((int)DirectX.OffScreenVRAM, Inis.graphics, "OffScreenVRAM");
	DirectX.OffScreenVRAM = i ? 1 : 0;
	i = get_config_int((int)DirectX.BackBufferVRAM, Inis.graphics, "BackBufferVRAM");
	DirectX.BackBufferVRAM = i ? 1 : 0;
	Options.GFXengine2 = (unsigned char)get_config_int((int)Options.GFXengine2, Inis.graphics, "GFXengine2");
	Settings.SixteenBit = (bool8)get_config_int((int)Settings.SixteenBit, Inis.graphics, "RenderSixteenBit");

	i = get_config_int((int)Settings.ForceTransparency, Inis.graphics, "Transparency");
	if (i == 0) {
		Settings.ForceTransparency = 0;
		Settings.ForceNoTransparency = 1;
	}
	else {
		Settings.ForceTransparency = 1;
		Settings.ForceNoTransparency = 0;
	}

	Settings.BGLayering = (bool8)get_config_int((int)Settings.BGLayering, Inis.graphics, "BGLayering");
	Settings.DisableGraphicWindows = (bool8)get_config_int((int)Settings.DisableGraphicWindows, Inis.graphics, "DisableGraphicWindows");
	Settings.DisableHDMA = (bool8)get_config_int((int)Settings.DisableHDMA, Inis.graphics, "DisableHDMA");
	Settings.Mode7Interpolate = (bool8)get_config_int((int)Settings.Mode7Interpolate, Inis.graphics, "Mode7Interpolate");
	GUI.SupportHiRes = (unsigned char)get_config_int((int)GUI.SupportHiRes, Inis.graphics, "hires");

	i = get_config_int((int)DirectX.wait_for_vsync, Inis.graphics, "wait_for_vsync");
	DirectX.wait_for_vsync = i ? 1 : 0;
	i = get_config_int((int)GUI.Stretch, Inis.graphics, "stretch");
	GUI.Stretch = (i >= 0 && i <= 2) ? i : GUI.Stretch;

	/* Read [sound] */
	GUI.DSuse = (BOOL)get_config_int((int)GUI.DSuse, Inis.sound, "DirectSound");
	GUI.DirectSoundNotify = (BOOL)get_config_int((int)GUI.DirectSoundNotify, Inis.sound, "DirectSoundNotify");
	Settings.NextAPUEnabled = (bool8)get_config_int((int)Settings.NextAPUEnabled, Inis.sound, "APUEnabled");
	i = get_config_int((int)GUI.SoundMode, Inis.sound, "SoundMode");
	GUI.SoundMode = (char)((0 <= i && i <= 23) ? i : GUI.SoundMode);
	i = get_config_int(Settings.InterpolatedSound, Inis.sound, "InterpolatedSound");
	Settings.InterpolatedSound = (0 <= i && i <= 3) ? i : Settings.InterpolatedSound;
	//Settings.LowPassFilter = (bool8)get_config_int((int)Settings.LowPassFilter, Inis.sound, "LowPassFilter");
	//Settings.SimpleFilter = (bool8)get_config_int((int)Settings.SimpleFilter, Inis.sound, "SimpleFilter");
	i = get_config_int((int)Settings.SoundBufferSize, Inis.sound, "SoundBufferLength");
	Settings.SoundBufferSize = (i >= 2 && i <= 9) ? i : Settings.SoundBufferSize;
	i = get_config_int(Settings.SoundMixInterval, Inis.sound, "SoundMixInterval");
	Settings.SoundMixInterval = (i >= 1 && i <= 999) ? i : Settings.SoundMixInterval;
	Settings.SoundEnvelopeHeightReading = (bool8)get_config_int((int)Settings.SoundEnvelopeHeightReading,
																Inis.sound,
																"SoundEnvelopeHeightReading");
	Settings.DisableSoundEcho = (bool8)get_config_int((int)Settings.DisableSoundEcho, Inis.sound, "DisableSoundEcho");
	Settings.DisableMasterVolume = (bool8)get_config_int((int)Settings.DisableMasterVolume, Inis.sound, "DisableMasterVolume");
	i = get_config_int(Options.VolumeLevel, Inis.sound, "VolumeLevel");
	Options.VolumeLevel = (i >= 1 && i <= 5) ? i : Options.VolumeLevel;
#if 0
	Settings.DisableSampleCaching = (bool8)get_config_int((int)Settings.DisableSampleCaching, Inis.sound, "DisableSampleCaching");
	i = get_config_int((int)Settings.AltSampleDecode, Inis.sound, "SampleDecodingMethod");
	Settings.AltSampleDecode = (uint8)((0 <= i && i <= 5) ? i : Settings.AltSampleDecode);
#endif
	Settings.ShutdownMaster = (bool8)!get_config_int((int)!Settings.ShutdownMaster, Inis.sound, "Shutdown");
	i = get_config_int((int)Settings.SoundSkipMethod, Inis.sound, "SoundSkipMethod");
	Settings.SoundSkipMethod = (uint8)((i >= 0 && i <= 3) ? i : Settings.SoundSkipMethod);
	Settings.SoundSync = (uint8)get_config_int((int)Settings.SoundSync, Inis.sound, "SoundSync");
	Settings.ReverseStereo = (uint8)get_config_int((int)Settings.ReverseStereo, Inis.sound, "ReverseStereo");
#if 0
	Settings.FixFrequency = (bool8)get_config_int((int)Settings.FixFrequency, Inis.sound, "FixFrequency");
#endif

	/* Read [cpu] */
	long H_Max_percentage = (long)get_config_int((int)Settings.CyclesPercentage, Inis.cpu, "CyclesPercentage");
	Settings.CyclesPercentage = (long)((H_Max_percentage > 0 && H_Max_percentage < 200) ? H_Max_percentage : Settings.CyclesPercentage);

	/* Read [timer] */
	i = get_config_int((int)(Settings.SkipFrames == AUTO_FRAMERATE ? AUTO_FRAMERATE : (Settings.SkipFrames - 1)), Inis.timer, "SkipFrames");
	if(i != AUTO_FRAMERATE && i >= 0 && i <= 9)
		Settings.SkipFrames = (uint32)i + 1;
	else
		Settings.SkipFrames = AUTO_FRAMERATE;
	i = get_config_int((int)Settings.AutoMaxSkipFrames, Inis.timer, "AutoMaxSkipFrames");
	Settings.AutoMaxSkipFrames = (uint32)((i >= 0 && i <= 10) ? i : Settings.AutoMaxSkipFrames);
	i = get_config_int((int)Settings.TurboSkipFrames, Inis.timer, "TurboSkipFrames");
	Settings.TurboSkipFrames = (uint32)((i >= 0 && i <= 15) ? i : Settings.TurboSkipFrames);
	Settings.FrameTimeNTSC = (uint32)get_config_int((int)Settings.FrameTimeNTSC, Inis.timer, "NTSCtiming");
	Settings.FrameTimePAL = (uint32)get_config_int((int)Settings.FrameTimePAL, Inis.timer, "PALtiming");

	/* Read [ControllerOption] */
	strcpy(tmpCharBuf, "JoystickConnected0");
	for(i = 0; i < 5; i++) {
		sprintf(&tmpCharBuf[17], "%d", i);
		Settings.JoystickConnected[i] = (bool8)get_config_int((int)Settings.JoystickConnected[i],
															  Inis.ControllerOption,
															  (const char *)tmpCharBuf);
	}

	Settings.Mouse = (uint8)get_config_int((int)Settings.Mouse, Inis.ControllerOption, "SnesMouse");
	Settings.SuperScope = (uint8)get_config_int((int)Settings.SuperScope, Inis.ControllerOption, "SnesSuperScope");
	Settings.MultiPlayer5 = (uint8)get_config_int((int)Settings.MultiPlayer5, Inis.ControllerOption, "SnesMultiPlayer5");
	Settings.Justifier = (uint8)get_config_int((int)Settings.Justifier, Inis.ControllerOption, "SnesJustifier");
	Settings.SecondJustifier = (uint8)get_config_int((int)Settings.SecondJustifier, Inis.ControllerOption, "SnesJustifier2");
	i = get_config_int((int)Settings.ControllerOption, Inis.ControllerOption, "DefaultPlugedInputDevice");
	if(i == 0 && Settings.MultiPlayer5)
		Settings.ControllerOption = SNES_MULTIPLAYER5;
	else if(i == 1)
		Settings.ControllerOption = SNES_JOYPAD;
	else if(i == 2 && Settings.Mouse)
		Settings.ControllerOption = SNES_MOUSE_SWAPPED;
	else if(i == 3 && Settings.Mouse)
		Settings.ControllerOption = SNES_MOUSE;
	else if(i == 4 && Settings.SuperScope)
		Settings.ControllerOption = SNES_SUPERSCOPE;
	else if(i == 5 && Settings.Justifier)
		Settings.ControllerOption = SNES_JUSTIFIER;
	else if(i == 6 && Settings.SecondJustifier)
		Settings.ControllerOption = SNES_JUSTIFIER_2;
	else
		Settings.ControllerOption = SNES_JOYPAD;

	i = get_config_int((int)(Options.JoyReadRateMask + 1), Inis.ControllerOption, "JoyReadRate");
	Options.JoyReadRateMask = (unsigned long)((i == 1 || i == 2 || i == 4 || i == 8) ? (i - 1) : Options.JoyReadRateMask);
	GUI.DIuse = (BOOL)get_config_int((int)GUI.DIuse, Inis.ControllerOption, "DirectInput");

	/* Read [joypadA1 2 3 4 5] */
	for(i = 0; i < 5; i++) {
		JoypadA[i].Left = (WORD)get_config_int((int)JoypadA[i].Left, Inis.joypadA[i], Key_button_str[0]);
		JoypadA[i].Right = (WORD)get_config_int((int)JoypadA[i].Right, Inis.joypadA[i], Key_button_str[1]);
		JoypadA[i].Up = (WORD)get_config_int((int)JoypadA[i].Up, Inis.joypadA[i], Key_button_str[2]);
		JoypadA[i].Down = (WORD)get_config_int((int)JoypadA[i].Down, Inis.joypadA[i], Key_button_str[3]);
		JoypadA[i].Start = (WORD)get_config_int((int)JoypadA[i].Start, Inis.joypadA[i], Key_button_str[4]);
		JoypadA[i].Select = (WORD)get_config_int((int)JoypadA[i].Select, Inis.joypadA[i], Key_button_str[5]);
		JoypadA[i].A = (WORD)get_config_int((int)JoypadA[i].A, Inis.joypadA[i], Key_button_str[6]);
		JoypadA[i].B = (WORD)get_config_int((int)JoypadA[i].B, Inis.joypadA[i], Key_button_str[7]);
		JoypadA[i].X = (WORD)get_config_int((int)JoypadA[i].X, Inis.joypadA[i], Key_button_str[8]);
		JoypadA[i].Y = (WORD)get_config_int((int)JoypadA[i].Y, Inis.joypadA[i], Key_button_str[9]);
		JoypadA[i].L = (WORD)get_config_int((int)JoypadA[i].L, Inis.joypadA[i], Key_button_str[10]);
		JoypadA[i].R = (WORD)get_config_int((int)JoypadA[i].R, Inis.joypadA[i], Key_button_str[11]);
		JoypadA[i].Turbo_Left = (get_config_int((int)JoypadA[i].Turbo_Left, Inis.joypadA[i], Key_button_str[12]) ? TRUE : FALSE);
		JoypadA[i].Turbo_Right = (get_config_int((int)JoypadA[i].Turbo_Right, Inis.joypadA[i], Key_button_str[13]) ? TRUE : FALSE);
		JoypadA[i].Turbo_Up = (get_config_int((int)JoypadA[i].Turbo_Up, Inis.joypadA[i], Key_button_str[14]) ? TRUE : FALSE);
		JoypadA[i].Turbo_Down = (get_config_int((int)JoypadA[i].Turbo_Down, Inis.joypadA[i], Key_button_str[15]) ? TRUE : FALSE);
		JoypadA[i].Turbo_Start = (get_config_int((int)JoypadA[i].Turbo_Start, Inis.joypadA[i], Key_button_str[16]) ? TRUE : FALSE);
		JoypadA[i].Turbo_Select = (get_config_int((int)JoypadA[i].Turbo_Select, Inis.joypadA[i], Key_button_str[17]) ? TRUE : FALSE);
		JoypadA[i].Turbo_A = (get_config_int((int)JoypadA[i].Turbo_A, Inis.joypadA[i], Key_button_str[18]) ? TRUE : FALSE);
		JoypadA[i].Turbo_B = (get_config_int((int)JoypadA[i].Turbo_B, Inis.joypadA[i], Key_button_str[19]) ? TRUE : FALSE);
		JoypadA[i].Turbo_X = (get_config_int((int)JoypadA[i].Turbo_X, Inis.joypadA[i], Key_button_str[20]) ? TRUE : FALSE);
		JoypadA[i].Turbo_Y = (get_config_int((int)JoypadA[i].Turbo_Y, Inis.joypadA[i], Key_button_str[21]) ? TRUE : FALSE);
		JoypadA[i].Turbo_L = (get_config_int((int)JoypadA[i].Turbo_L, Inis.joypadA[i], Key_button_str[22]) ? TRUE : FALSE);
		JoypadA[i].Turbo_R = (get_config_int((int)JoypadA[i].Turbo_R, Inis.joypadA[i], Key_button_str[23]) ? TRUE : FALSE);
	}

	/* Read [FunctionCodeA] */
	strcpy(tmpCharBuf, "Code0");
	for(i = 0; i < 8; i++) {
		sprintf(&tmpCharBuf[4], "%d", i);
		FunctionCodeA[i] = (DWORD)get_config_int((int)FunctionCodeA[i], Inis.functioncodeA, (const char *)tmpCharBuf);
	}

	/* Read [joypadB1 2 3 4 5] */
	for(i = 0; i < 5; i++) {
		JoypadB[i].Left = (WORD)get_config_int((int)JoypadB[i].Left, Inis.joypadB[i], Key_button_str[0]);
		JoypadB[i].Right = (WORD)get_config_int((int)JoypadB[i].Right, Inis.joypadB[i], Key_button_str[1]);
		JoypadB[i].Up = (WORD)get_config_int((int)JoypadB[i].Up, Inis.joypadB[i], Key_button_str[2]);
		JoypadB[i].Down = (WORD)get_config_int((int)JoypadB[i].Down, Inis.joypadB[i], Key_button_str[3]);
		JoypadB[i].Start = (WORD)get_config_int((int)JoypadB[i].Start, Inis.joypadB[i], Key_button_str[4]);
		JoypadB[i].Select = (WORD)get_config_int((int)JoypadB[i].Select, Inis.joypadB[i], Key_button_str[5]);
		JoypadB[i].A = (WORD)get_config_int((int)JoypadB[i].A, Inis.joypadB[i], Key_button_str[6]);
		JoypadB[i].B = (WORD)get_config_int((int)JoypadB[i].B, Inis.joypadB[i], Key_button_str[7]);
		JoypadB[i].X = (WORD)get_config_int((int)JoypadB[i].X, Inis.joypadB[i], Key_button_str[8]);
		JoypadB[i].Y = (WORD)get_config_int((int)JoypadB[i].Y, Inis.joypadB[i], Key_button_str[9]);
		JoypadB[i].L = (WORD)get_config_int((int)JoypadB[i].L, Inis.joypadB[i], Key_button_str[10]);
		JoypadB[i].R = (WORD)get_config_int((int)JoypadB[i].R, Inis.joypadB[i], Key_button_str[11]);
		JoypadB[i].Turbo_Left = (get_config_int((int)JoypadB[i].Turbo_Left, Inis.joypadB[i], Key_button_str[12]) ? TRUE : FALSE);
		JoypadB[i].Turbo_Right = (get_config_int((int)JoypadB[i].Turbo_Right, Inis.joypadB[i], Key_button_str[13]) ? TRUE : FALSE);
		JoypadB[i].Turbo_Up = (get_config_int((int)JoypadB[i].Turbo_Up, Inis.joypadB[i], Key_button_str[14]) ? TRUE : FALSE);
		JoypadB[i].Turbo_Down = (get_config_int((int)JoypadB[i].Turbo_Down, Inis.joypadB[i], Key_button_str[15]) ? TRUE : FALSE);
		JoypadB[i].Turbo_Start = (get_config_int((int)JoypadB[i].Turbo_Start, Inis.joypadB[i], Key_button_str[16]) ? TRUE : FALSE);
		JoypadB[i].Turbo_Select = (get_config_int((int)JoypadB[i].Turbo_Select, Inis.joypadB[i], Key_button_str[17]) ? TRUE : FALSE);
		JoypadB[i].Turbo_A = (get_config_int((int)JoypadB[i].Turbo_A, Inis.joypadB[i], Key_button_str[18]) ? TRUE : FALSE);
		JoypadB[i].Turbo_B = (get_config_int((int)JoypadB[i].Turbo_B, Inis.joypadB[i], Key_button_str[19]) ? TRUE : FALSE);
		JoypadB[i].Turbo_X = (get_config_int((int)JoypadB[i].Turbo_X, Inis.joypadB[i], Key_button_str[20]) ? TRUE : FALSE);
		JoypadB[i].Turbo_Y = (get_config_int((int)JoypadB[i].Turbo_Y, Inis.joypadB[i], Key_button_str[21]) ? TRUE : FALSE);
		JoypadB[i].Turbo_L = (get_config_int((int)JoypadB[i].Turbo_L, Inis.joypadB[i], Key_button_str[22]) ? TRUE : FALSE);
		JoypadB[i].Turbo_R = (get_config_int((int)JoypadB[i].Turbo_R, Inis.joypadB[i], Key_button_str[23]) ? TRUE : FALSE);
	}

	/* Read [FunctionCodeB] */
	strcpy(tmpCharBuf, "Code0");
	for(i = 0; i < 8; i++) {
		sprintf(&tmpCharBuf[4], "%d", i);
		FunctionCodeB[i] = (DWORD)get_config_int((int)FunctionCodeB[i], Inis.functioncodeB, (const char *)tmpCharBuf);
	}

	/* Read [gui] */
	tmpcharptr = get_config_string(Inis.gui, "LastROMFileName");
	if(tmpcharptr[0]) {
		strcpy(ROMFileName, tmpcharptr);
	}
	mbcs_splitpath(ROMFileName, drive, dir, fname, ext);
	mbcs_makepath(InitialROMFileDirectory, drive, dir, NULL, NULL);

	strcpy(tmpCharBuf, "RecentROMFileName1");
	for(i = 0; i < 10; i++) {
		sprintf(&tmpCharBuf[17], "%X", i + 1);
		tmpcharptr = get_config_string(Inis.gui, (const char *)tmpCharBuf);
		if(tmpcharptr[0] && strlen(tmpcharptr) < _MAX_PATH)
			strcpy(&RecentROMFileName[_MAX_PATH * i], tmpcharptr);
		else
			RecentROMFileName[_MAX_PATH * i] = 0;
	}

	tmpcharptr = get_config_string(Inis.gui, "SnapShotDirectory");
	if(tmpcharptr[0] && strlen((const char *)tmpcharptr) < _MAX_PATH)
		strcpy(tmpCharBuf, (const char *)tmpcharptr);
	else
		strcpy(tmpCharBuf, (const char *)GUI.FreezeFileDir);

	if(access((const char *)tmpCharBuf, 0) == 0)
		strcpy(GUI.FreezeFileDir, (const char *)tmpCharBuf);
	mbcs_splitpath(GUI.FreezeFileDir, drive, dir, fname, ext);
	if(dir[0] && fname[0] == 0) {
		dir[strlen(dir) - 1] = 0;
	}
	mbcs_makepath(GUI.FreezeFileDir, drive, dir, fname, ext);
	
	tmpcharptr = get_config_string(Inis.gui, "ScreenShotDirectory");
	if(tmpcharptr[0] && strlen((const char *)tmpcharptr) < _MAX_PATH)
		strcpy(tmpCharBuf, (const char *)tmpcharptr);
	else
		strcpy(tmpCharBuf, (const char *)Memory.ScreenShotDirectory);
	
	if(access((const char *)tmpCharBuf, 0) == 0)
		strcpy(Memory.ScreenShotDirectory, (const char *)tmpCharBuf);
	mbcs_splitpath(Memory.ScreenShotDirectory, drive, dir, fname, ext);
	if(dir[0] && fname[0] == 0) {
		dir[strlen(dir) - 1] = 0;
	}
	mbcs_makepath(Memory.ScreenShotDirectory, drive, dir, fname, ext);

	tmpcharptr = get_config_string(Inis.gui, "SPCDumpDirectory");
	if(tmpcharptr[0] && strlen((const char *)tmpcharptr) < _MAX_PATH)
		strcpy(tmpCharBuf, (const char *)tmpcharptr);
	else
		strcpy(tmpCharBuf, (const char *)Memory.SPCDumpDirectory);
	
	if(access((const char *)tmpCharBuf, 0) == 0)
		strcpy(Memory.SPCDumpDirectory, (const char *)tmpCharBuf);
	mbcs_splitpath(Memory.SPCDumpDirectory, drive, dir, fname, ext);
	if(dir[0] && fname[0] == 0) {
		dir[strlen(dir) - 1] = 0;
	}
	mbcs_makepath(Memory.SPCDumpDirectory, drive, dir, fname, ext);

	i = get_config_int(GUI.Language, Inis.gui, "Language");
	GUI.Language = (BYTE)((i == 1 || i == 0) ? i : GUI.Language);

	GUI.WindowRectSetting.left = get_config_int(-0x7fffffff, Inis.gui, "WindowRectLeft");
	GUI.WindowRectSetting.right = get_config_int(-0x7fffffff, Inis.gui, "WindowRectRight");
	GUI.WindowRectSetting.top = get_config_int(-0x7fffffff, Inis.gui, "WindowRectTop");
	GUI.WindowRectSetting.bottom = get_config_int(-0x7fffffff, Inis.gui, "WindowRectBottom");
	if(GUI.WindowRectSetting.left == -0x7fffffff ||
	   GUI.WindowRectSetting.right == -0x7fffffff ||
	   GUI.WindowRectSetting.top == -0x7fffffff ||
	   GUI.WindowRectSetting.bottom == -0x7fffffff)
		GUI.WindowRectSetting.left = GUI.WindowRectSetting.right = GUI.WindowRectSetting.top = GUI.WindowRectSetting.bottom = -1;
	else {
		if((GUI.WindowRectSetting.right - GUI.WindowRectSetting.left) < GUI.DefaultWindowSizeW) {
			GUI.WindowRectSetting.right = (GUI.WindowRectSetting.left + GUI.DefaultWindowSizeW);
		}
		if((GUI.WindowRectSetting.bottom - GUI.WindowRectSetting.top) < GUI.DefaultWindowSizeH) {
			GUI.WindowRectSetting.bottom = (GUI.WindowRectSetting.top + GUI.DefaultWindowSizeH);
			GUI.NeedWindowSizeAddition = 2;
		}
	}

	GUI.FileBrowseWindow = (unsigned char)get_config_int((int)GUI.FileBrowseWindow, Inis.gui, "FileBrowseWindow");
	
	uosxplorerposition.WindowRect.left = get_config_int(uosxplorerposition.WindowRect.left, Inis.gui, "uosXpWindowRectLeft");
	uosxplorerposition.WindowRect.right = get_config_int(uosxplorerposition.WindowRect.right, Inis.gui, "uosXpWindowRectRight");
	uosxplorerposition.WindowRect.top = get_config_int(uosxplorerposition.WindowRect.top, Inis.gui, "uosXpWindowRectTop");
	uosxplorerposition.WindowRect.bottom = get_config_int(uosxplorerposition.WindowRect.bottom, Inis.gui, "uosXpWindowRectBottom");
	uosxplorerposition.Maximize = (BOOL)get_config_int(uosxplorerposition.Maximize, Inis.gui, "uosXpMaximize");
	uosxplorerposition.VSplitBarPos0 = get_config_int(uosxplorerposition.VSplitBarPos0, Inis.gui, "uosXpVSplitBarPos0");
	uosxplorerposition.ColumnWidth0 = get_config_int(uosxplorerposition.ColumnWidth0, Inis.gui, "uosXpColumnWidth0");
	uosxplorerposition.ColumnWidth1 = get_config_int(uosxplorerposition.ColumnWidth1, Inis.gui, "uosXpColumnWidth1");
	uosxplorerposition.ColumnWidth2 = get_config_int(uosxplorerposition.ColumnWidth2, Inis.gui, "uosXpColumnWidth2");
	uosxplorerposition.ColumnWidth3 = get_config_int(uosxplorerposition.ColumnWidth3, Inis.gui, "uosXpColumnWidth3");

	/* Read [screenshot] */
	i = get_config_int((int)Options.SaveImagefileFormat, Inis.screenshot, "SaveImagefileFormat");
	Options.SaveImagefileFormat = (uint8)((0 <= i && i <= 3) ? i : Options.SaveImagefileFormat);
	Options.ReduceTheColorTo256IfWithin = (uint8)get_config_int((int)Options.ReduceTheColorTo256IfWithin,
																Inis.screenshot,
																"ReduceTheColorTo256IfWithin");
	Options.ImagefileHiresStretch = (bool8)get_config_int((int)Options.ImagefileHiresStretch,
														  Inis.screenshot,
														  "ImagefileHiresStretch");
	Options.RLEcompression256bitmap = (bool8)get_config_int((int)Options.RLEcompression256bitmap,
															Inis.screenshot,
															"RLEcompression256bitmap");

	/* Read [BIOS] */
	tmpcharptr = get_config_string(Inis.bios, "BSX");
	if(tmpcharptr[0] && strlen((const char *)tmpcharptr) < _MAX_PATH)
		strcpy(Memory.BSXBIOSROMFileName, (const char *)tmpcharptr);
	tmpcharptr = get_config_string(Inis.bios, "SFT");
	if(tmpcharptr[0] && strlen((const char *)tmpcharptr) < _MAX_PATH)
		strcpy(Memory.SFTBIOSROMFileName, (const char *)tmpcharptr);

	/* Read [debug] */
	Options.DisplayScreenBufferAll = (bool8)get_config_int((int)Options.DisplayScreenBufferAll,
														   Inis.debug,
														   "DisplayScreenBufferAll");

	/* Read [option] */
	Settings.ApplyCheats = (bool8)get_config_int((int)Settings.ApplyCheats, Inis.option, "ApplyCheats");
	Settings.DisplayFrameRate = (bool8)get_config_int((int)Settings.DisplayFrameRate, Inis.option, "ShowFPS");
	Settings.Mute = (bool8)get_config_int((int)Settings.Mute, Inis.option, "MuteSound");
	Settings.AutoSaveDelay = (int32)get_config_int((int)Settings.AutoSaveDelay, Inis.option, "AutoSaveDelay");
	Xxxsystem.NormalSnap = (bool8)get_config_int((int)Xxxsystem.NormalSnap, Inis.option, "NormalSnap");
	Settings.SDD1Pack = (bool8)get_config_int((int)Settings.SDD1Pack, Inis.option, "SDD1Pack");
	Settings.DisableIRQCycleCount = (bool8)get_config_int((int)Settings.DisableIRQCycleCount, Inis.option, "DisableIRQCycleCount");
	Xxxcdda.Enable = (bool8)get_config_int((int)Xxxcdda.Enable, Inis.option, "UseCDDA");
	i = get_config_int((int)Xxxcdda.Drv, Inis.option, "CDDrv");
	Xxxcdda.Drv = (uint32)((0 <= i && i <= 26) ? i : Xxxcdda.Drv);

	/* check RevID */
	if(RevID != get_config_int(0, Inis.option, "RevID")) {
		remove((const char *)cfgfilename);
		if(!load_configfile(cfgfilename,
							config_data,
							write_config)) {
			if(!configfile_cannot_create) {
				write_config();
				flush_configfile();
			}
		}
	}

	return;
}

static void write_config()
{
	if(configfile_cannot_create)
		return;
	
	int i, c;
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fname [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];
	char tmpCharBuf[_MAX_PATH + 1];

	/* Write [windows] */
	switch(GUI.PriorityClass) {
	case IDLE_PRIORITY_CLASS: i = 0; break;
	case NORMAL_PRIORITY_CLASS: i = 1; break;
	case HIGH_PRIORITY_CLASS: i = 2; break;
	case REALTIME_PRIORITY_CLASS: i = 3; break;
	default: i = 1;  break;
	}
	set_config_int(Inis.windows, "PriorityClass", i);

	switch(GUI.ThreadPriority) {
	case THREAD_PRIORITY_LOWEST: i = 0; break;
	case THREAD_PRIORITY_BELOW_NORMAL: i = 1; break;
	case THREAD_PRIORITY_NORMAL: i = 2; break;
	case THREAD_PRIORITY_ABOVE_NORMAL: i = 3; break;
	case THREAD_PRIORITY_HIGHEST: i = 4; break;
	case THREAD_PRIORITY_IDLE: i = 5; break;
	case THREAD_PRIORITY_TIME_CRITICAL: i = 6; break;
	default: i = 2; break;
	}
	set_config_int(Inis.windows, "ThreadPriority", i);

	set_config_int(Inis.windows, "WindowTopMost", (int)GUI.WindowTopMost);

	/* Write [graphics] */
	set_config_int(Inis.graphics, "Fullscreen", (int)GUI.FullScreen);
	set_config_int(Inis.graphics, "DoubleRendering", (int)GUI.DoubleRendering);
	set_config_int(Inis.graphics, "FullscreenDisplayModeIndexNormal16", (int)GUI.FullscreenDisplayModeIndexNormal16);
	set_config_int(Inis.graphics, "FullscreenDisplayModeIndexDouble16", (int)GUI.FullscreenDisplayModeIndexDouble16);
	set_config_int(Inis.graphics, "FullscreenDisplayModeIndexNormal8", (int)GUI.FullscreenDisplayModeIndexNormal8);
	set_config_int(Inis.graphics, "FullscreenDisplayModeIndexDouble8", (int)GUI.FullscreenDisplayModeIndexDouble8);
	set_config_int(Inis.graphics, "DoubleRenderingModes16", (int)GUI.DoubleRenderingModes16);
	set_config_int(Inis.graphics, "DoubleRenderingModes8", (int)GUI.DoubleRenderingModes8);
	set_config_int(Inis.graphics, "DirectScreenRendering", (int)GUI.dRendering);
	set_config_int(Inis.graphics, "DisableHardwareScale", (int)DirectX.DisableHardwareScale);
	set_config_int(Inis.graphics, "DoubleBuffering", (int)GUI.DoubleBuffered);
	set_config_int(Inis.graphics, "ForceWindowedDirectDrawClipper", (int)GUI.ForceWindowedDirectDrawClipper);
	set_config_int(Inis.graphics, "OffScreenVRAM", (int)DirectX.OffScreenVRAM);
	set_config_int(Inis.graphics, "BackBufferVRAM", (int)DirectX.BackBufferVRAM);
	set_config_int(Inis.graphics, "GFXengine2", (int)Options.GFXengine2);
	set_config_int(Inis.graphics, "wait_for_vsync", (int)DirectX.wait_for_vsync);
	set_config_int(Inis.graphics, "stretch", (int)GUI.Stretch);
	set_config_int(Inis.graphics, "RenderSixteenBit", (int)Settings.SixteenBit);
	set_config_int(Inis.graphics, "Transparency", (int)Settings.Transparency);
	set_config_int(Inis.graphics, "BGLayering", (int)Settings.BGLayering);
	set_config_int(Inis.graphics, "DisableGraphicWindows", (int)Settings.DisableGraphicWindows);
	set_config_int(Inis.graphics, "DisableHDMA", (int)Settings.DisableHDMA);
	set_config_int(Inis.graphics, "hires", (int)GUI.SupportHiRes);
	set_config_int(Inis.graphics, "Mode7Interpolate", (int)Settings.Mode7Interpolate);


	/* Write [sound] */
	set_config_int(Inis.sound, "DirectSound", (int)GUI.DSuse);
	set_config_int(Inis.sound, "DirectSoundNotify", (int)GUI.DirectSoundNotify);
	set_config_int(Inis.sound, "APUEnabled", (int)Settings.NextAPUEnabled);
	set_config_int(Inis.sound, "SoundMode", (int)GUI.SoundMode);
	set_config_int(Inis.sound, "InterpolatedSound", Settings.InterpolatedSound);
	set_config_int(Inis.sound, "LowPassFilter", (int)Settings.LowPassFilter);
	set_config_int(Inis.sound, "SimpleFilter", (int)Settings.SimpleFilter);
	set_config_int(Inis.sound, "SoundBufferLength", (int)Settings.SoundBufferSize);
	set_config_int(Inis.sound, "SoundMixInterval", (int)Settings.SoundMixInterval);
	set_config_int(Inis.sound, "SoundEnvelopeHeightReading", (int)Settings.SoundEnvelopeHeightReading);
	set_config_int(Inis.sound, "DisableSoundEcho", (int)Settings.DisableSoundEcho);
	set_config_int(Inis.sound, "DisableMasterVolume", (int)Settings.DisableMasterVolume);
	set_config_int(Inis.sound, "VolumeLevel", Options.VolumeLevel);
#if 0
	set_config_int(Inis.sound, "DisableSampleCaching", (int)Settings.DisableSampleCaching);
	set_config_int(Inis.sound, "SampleDecodingMethod", (int)Settings.AltSampleDecode);
#endif
	set_config_int(Inis.sound, "Shutdown", (int)(Settings.ShutdownMaster ? 0 : 1));
	set_config_int(Inis.sound, "SoundSkipMethod", (int)Settings.SoundSkipMethod);
	set_config_int(Inis.sound, "SoundSync", (int)Settings.SoundSync);
	set_config_int(Inis.sound, "ReverseStereo", (int)Settings.ReverseStereo);
#if 0
	set_config_int(Inis.sound, "FixFrequency", (int)Options.FixFrequency);
#endif

	/* Write [cpu] */
	set_config_int(Inis.cpu, "CyclesPercentage", (int)Settings.CyclesPercentage);


	/* Write [timer] */
	if (Settings.SkipFrames != AUTO_FRAMERATE)
		set_config_int(Inis.timer, "SkipFrames", (int)(Settings.SkipFrames - 1));
	else
		set_config_string(Inis.timer, "SkipFrames", "");
	set_config_int(Inis.timer, "AutoMaxSkipFrames", (int)Settings.AutoMaxSkipFrames);
	set_config_int(Inis.timer, "TurboSkipFrames", (int)Settings.TurboSkipFrames);
	set_config_int(Inis.timer, "NTSCtiming", (int)Settings.FrameTimeNTSC);
	set_config_int(Inis.timer, "PALtiming", (int)Settings.FrameTimePAL);


	/* Write [ControllerOption] */
	strcpy(tmpCharBuf, "JoystickConnected0");
	for(i = 0; i < 5; i++) {
		sprintf(&tmpCharBuf[17], "%d", i);
		set_config_int(Inis.ControllerOption, (const char *)tmpCharBuf, (int)Settings.JoystickConnected[i]);
	}

	set_config_int(Inis.ControllerOption, "SnesMouse", (int)Settings.Mouse);
	set_config_int(Inis.ControllerOption, "SnesSuperScope", (int)Settings.SuperScope);
	set_config_int(Inis.ControllerOption, "SnesMultiPlayer5", (int)Settings.MultiPlayer5);
	set_config_int(Inis.ControllerOption, "SnesJustifier", (int)Settings.Justifier);
	set_config_int(Inis.ControllerOption, "SnesJustifier2", (int)Settings.SecondJustifier);


	switch(Settings.ControllerOption){
	case SNES_MULTIPLAYER5: i = 0; break;
	case SNES_JOYPAD: i = 1; break;
	case SNES_MOUSE_SWAPPED: i = 2; break;
	case SNES_MOUSE: i = 3; break;
	case SNES_SUPERSCOPE: i = 4; break;
	case SNES_JUSTIFIER: i = 5; break;
	case SNES_JUSTIFIER_2: i = 6; break;
	default: i = 1; break;
	}
	set_config_int(Inis.ControllerOption, "DefaultPlugedInputDevice", i);
	set_config_int(Inis.ControllerOption, "JoyReadRate", (int)(Options.JoyReadRateMask + 1));
	set_config_int(Inis.ControllerOption, "DirectInput", (int)GUI.DIuse);

	/* Write [joypadA1 2 3 4 5] */
	for(i = 0; i < 5; i++) {
		set_config_int(Inis.joypadA[i], Key_button_str[0], (int)JoypadA[i].Left);
		set_config_int(Inis.joypadA[i], Key_button_str[1], (int)JoypadA[i].Right);
		set_config_int(Inis.joypadA[i], Key_button_str[2], (int)JoypadA[i].Up);
		set_config_int(Inis.joypadA[i], Key_button_str[3], (int)JoypadA[i].Down);
		set_config_int(Inis.joypadA[i], Key_button_str[4], (int)JoypadA[i].Start);
		set_config_int(Inis.joypadA[i], Key_button_str[5], (int)JoypadA[i].Select);
		set_config_int(Inis.joypadA[i], Key_button_str[6], (int)JoypadA[i].A);
		set_config_int(Inis.joypadA[i], Key_button_str[7], (int)JoypadA[i].B);
		set_config_int(Inis.joypadA[i], Key_button_str[8], (int)JoypadA[i].X);
		set_config_int(Inis.joypadA[i], Key_button_str[9], (int)JoypadA[i].Y);
		set_config_int(Inis.joypadA[i], Key_button_str[10], (int)JoypadA[i].L);
		set_config_int(Inis.joypadA[i], Key_button_str[11], (int)JoypadA[i].R);
		set_config_int(Inis.joypadA[i], Key_button_str[12], (int)JoypadA[i].Turbo_Left);
		set_config_int(Inis.joypadA[i], Key_button_str[13], (int)JoypadA[i].Turbo_Right);
		set_config_int(Inis.joypadA[i], Key_button_str[14], (int)JoypadA[i].Turbo_Up);
		set_config_int(Inis.joypadA[i], Key_button_str[15], (int)JoypadA[i].Turbo_Down);
		set_config_int(Inis.joypadA[i], Key_button_str[16], (int)JoypadA[i].Turbo_Start);
		set_config_int(Inis.joypadA[i], Key_button_str[17], (int)JoypadA[i].Turbo_Select);
		set_config_int(Inis.joypadA[i], Key_button_str[18], (int)JoypadA[i].Turbo_A);
		set_config_int(Inis.joypadA[i], Key_button_str[19], (int)JoypadA[i].Turbo_B);
		set_config_int(Inis.joypadA[i], Key_button_str[20], (int)JoypadA[i].Turbo_X);
		set_config_int(Inis.joypadA[i], Key_button_str[21], (int)JoypadA[i].Turbo_Y);
		set_config_int(Inis.joypadA[i], Key_button_str[22], (int)JoypadA[i].Turbo_L);
		set_config_int(Inis.joypadA[i], Key_button_str[23], (int)JoypadA[i].Turbo_R);
	}

	/* Write [FunctionCodeA] */
	strcpy(tmpCharBuf, "Code0");
	for(i = 0; i < 8; i++) {
		sprintf(&tmpCharBuf[4], "%d", i);
		set_config_int(Inis.functioncodeA, (const char *)tmpCharBuf, (int)FunctionCodeA[i]);
	}

	/* Write [joypadB1 2 3 4 5] */
	for(i = 0; i < 5; i++) {
		set_config_int(Inis.joypadB[i], Key_button_str[0], (int)JoypadB[i].Left);
		set_config_int(Inis.joypadB[i], Key_button_str[1], (int)JoypadB[i].Right);
		set_config_int(Inis.joypadB[i], Key_button_str[2], (int)JoypadB[i].Up);
		set_config_int(Inis.joypadB[i], Key_button_str[3], (int)JoypadB[i].Down);
		set_config_int(Inis.joypadB[i], Key_button_str[4], (int)JoypadB[i].Start);
		set_config_int(Inis.joypadB[i], Key_button_str[5], (int)JoypadB[i].Select);
		set_config_int(Inis.joypadB[i], Key_button_str[6], (int)JoypadB[i].A);
		set_config_int(Inis.joypadB[i], Key_button_str[7], (int)JoypadB[i].B);
		set_config_int(Inis.joypadB[i], Key_button_str[8], (int)JoypadB[i].X);
		set_config_int(Inis.joypadB[i], Key_button_str[9], (int)JoypadB[i].Y);
		set_config_int(Inis.joypadB[i], Key_button_str[10], (int)JoypadB[i].L);
		set_config_int(Inis.joypadB[i], Key_button_str[11], (int)JoypadB[i].R);
		set_config_int(Inis.joypadB[i], Key_button_str[12], (int)JoypadB[i].Turbo_Left);
		set_config_int(Inis.joypadB[i], Key_button_str[13], (int)JoypadB[i].Turbo_Right);
		set_config_int(Inis.joypadB[i], Key_button_str[14], (int)JoypadB[i].Turbo_Up);
		set_config_int(Inis.joypadB[i], Key_button_str[15], (int)JoypadB[i].Turbo_Down);
		set_config_int(Inis.joypadB[i], Key_button_str[16], (int)JoypadB[i].Turbo_Start);
		set_config_int(Inis.joypadB[i], Key_button_str[17], (int)JoypadB[i].Turbo_Select);
		set_config_int(Inis.joypadB[i], Key_button_str[18], (int)JoypadB[i].Turbo_A);
		set_config_int(Inis.joypadB[i], Key_button_str[19], (int)JoypadB[i].Turbo_B);
		set_config_int(Inis.joypadB[i], Key_button_str[20], (int)JoypadB[i].Turbo_X);
		set_config_int(Inis.joypadB[i], Key_button_str[21], (int)JoypadB[i].Turbo_Y);
		set_config_int(Inis.joypadB[i], Key_button_str[22], (int)JoypadB[i].Turbo_L);
		set_config_int(Inis.joypadB[i], Key_button_str[23], (int)JoypadB[i].Turbo_R);
	}

	/* Write [FunctionCodeB] */
	strcpy(tmpCharBuf, "Code0");
	for(i = 0; i < 8; i++) {
		sprintf(&tmpCharBuf[4], "%d", i);
		set_config_int(Inis.functioncodeB, (const char *)tmpCharBuf, (int)FunctionCodeB[i]);
	}

	/* Write [gui] */
	set_config_string(Inis.gui, "LastROMFileName", ROMFileName);
	strcpy(tmpCharBuf, "RecentROMFileName1");
	for(i = 0; i < 10; i++) {
		sprintf(&tmpCharBuf[17], "%X", i + 1);
		set_config_string(Inis.gui, (const char *)tmpCharBuf, &RecentROMFileName[_MAX_PATH * i]);
	}

	mbcs_splitpath(GUI.FreezeFileDir, drive, dir, fname, ext);
	if(dir[0] == 0)
		mbcs_makepath (GUI.FreezeFileDir, drive, "\\", NULL, NULL);
	else
		mbcs_makepath (GUI.FreezeFileDir, drive, dir, fname, ext);
	set_config_string(Inis.gui, "SnapShotDirectory", GUI.FreezeFileDir);

	mbcs_splitpath(Memory.ScreenShotDirectory, drive, dir, fname, ext);
	if(dir[0] == 0)
		mbcs_makepath (Memory.ScreenShotDirectory, drive, "\\", NULL, NULL);
	else
		mbcs_makepath (Memory.ScreenShotDirectory, drive, dir, fname, ext);
	set_config_string(Inis.gui, "ScreenShotDirectory", Memory.ScreenShotDirectory);

	mbcs_splitpath(Memory.SPCDumpDirectory, drive, dir, fname, ext);
	if(dir[0] == 0)
		mbcs_makepath (Memory.SPCDumpDirectory, drive, "\\", NULL, NULL);
	else
		mbcs_makepath (Memory.SPCDumpDirectory, drive, dir, fname, ext);
	set_config_string(Inis.gui, "SPCDumpDirectory", Memory.SPCDumpDirectory);

	set_config_int(Inis.gui, "Language", (int)GUI.Language);

	set_config_int(Inis.gui, "WindowRectLeft", (int)GUI.WindowRectSetting.left);
	set_config_int(Inis.gui, "WindowRectRight", (int)GUI.WindowRectSetting.right);
	set_config_int(Inis.gui, "WindowRectTop", (int)GUI.WindowRectSetting.top);
	set_config_int(Inis.gui, "WindowRectBottom", (int)GUI.WindowRectSetting.bottom);

	set_config_int(Inis.gui, "FileBrowseWindow", (int)GUI.FileBrowseWindow);

	set_config_int(Inis.gui, "uosXpWindowRectLeft", (int)uosxplorerposition.WindowRect.left);
	set_config_int(Inis.gui, "uosXpWindowRectRight", (int)uosxplorerposition.WindowRect.right);
	set_config_int(Inis.gui, "uosXpWindowRectTop", (int)uosxplorerposition.WindowRect.top);
	set_config_int(Inis.gui, "uosXpWindowRectBottom", (int)uosxplorerposition.WindowRect.bottom);
	set_config_int(Inis.gui, "uosXpMaximize", (int)uosxplorerposition.Maximize);
	set_config_int(Inis.gui, "uosXpVSplitBarPos0", (int)uosxplorerposition.VSplitBarPos0);
	set_config_int(Inis.gui, "uosXpColumnWidth0", (int)uosxplorerposition.ColumnWidth0);
	set_config_int(Inis.gui, "uosXpColumnWidth1", (int)uosxplorerposition.ColumnWidth1);
	set_config_int(Inis.gui, "uosXpColumnWidth2", (int)uosxplorerposition.ColumnWidth2);
	set_config_int(Inis.gui, "uosXpColumnWidth3", (int)uosxplorerposition.ColumnWidth3);


	/* Write [screenshot] */
	set_config_int(Inis.screenshot, "SaveImagefileFormat", (int)Options.SaveImagefileFormat);
	set_config_int(Inis.screenshot, "ReduceTheColorTo256", (int)Options.ReduceTheColorTo256);
	set_config_int(Inis.screenshot, "ReduceTheColorTo256IfWithin", (int)Options.ReduceTheColorTo256IfWithin);
	set_config_int(Inis.screenshot, "ImagefileHiresStretch", (int)Options.ImagefileHiresStretch);
	set_config_int(Inis.screenshot, "RLEcompression256bitmap", (int)Options.RLEcompression256bitmap);

	/* Write [BIOS] */
	set_config_string(Inis.bios, "BSX", Memory.BSXBIOSROMFileName);
	set_config_string(Inis.bios, "SFT", Memory.SFTBIOSROMFileName);

	/* Write [debug] */
	set_config_int(Inis.debug, "DisplayScreenBufferAll", (int)Options.DisplayScreenBufferAll);


	/* Write [option] */
	set_config_int(Inis.option, "ApplyCheats", (int)Settings.ApplyCheats);
	set_config_int(Inis.option, "ShowFPS", (int)Settings.DisplayFrameRate);
	set_config_int(Inis.option, "MuteSound", (int)Settings.Mute);
    set_config_int(Inis.option, "AutoSaveDelay", Settings.AutoSaveDelay);
    set_config_int(Inis.option, "NormalSnap", Xxxsystem.NormalSnap);
    set_config_int(Inis.option, "SDD1Pack", Settings.SDD1Pack);
    set_config_int(Inis.option, "DisableIRQCycleCount", Settings.DisableIRQCycleCount);
    set_config_int(Inis.option, "UseCDDA", Xxxcdda.Enable);
    set_config_int(Inis.option, "CDDrv", Xxxcdda.Drv);

    set_config_int(Inis.option, "RevID", RevID);

	return;
}
