#include "snes9x.h"
#include "memmap.h"
#include "debug.h"
#include "cpuexec.h"
#include "ppu.h"
#include "snapshot.h"
#include "apu.h"
#include "display.h"
#include "gfx.h"
#include "directx.h"
#include "render.h"
#include "commutil.h"
#include "uosnesx.h"
#include "uowinput.h"
#include "uowsound.h"
#include "AVIOutput.h"

#include <direct.h>
#include <io.h>

void S9xSetWinPixelFormat ();

typedef struct
{
    uint8 red;
    uint8 green;
    uint8 blue;
} Colour;

static Colour FixedColours [256];
extern "C" {
uint8 S9xwPalette [0x10000];
int color_diff [0x10000];
};

void S9xParseArg (char **argv, int &i, int argc)
{
#if 0
    if (strcasecmp (argv [i], "-restore") == 0)
    {
        WinDeleteRegistryEntries ();
        WinSetDefaultValues ();
	}
	if (strcasecmp (argv[i], "-hidemenu") == 0)
	{
		GUI.HideMenu = true;
	}
#endif
}

void S9xExtraUsage ()
{
}

void S9xTextMode( void)
{
}

void S9xGraphicsMode ()
{
}

void S9xExit( void)
{
	MessageBox(NULL, Languages[ GUI.Language].S9xExitMessage, "uosnesw", MB_OK | MB_ICONSTOP);
	SendMessage(GUI.hWnd, WM_COMMAND, ID_FILE_EXIT, 0);
}

extern "C" const char *S9xGetFilename (const char *e)
{
    static char filename [_MAX_PATH + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fname [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];

    if (strlen (GUI.FreezeFileDir))
    {
        mbcs_splitpath (Memory.ROMFilename, drive, dir, fname, ext);
        strcpy (filename, GUI.FreezeFileDir);
        strcat (filename, "\\");
        strcat (filename, fname);
        strcat (filename, e);
    }
    else
    {
        mbcs_splitpath (Memory.ROMFilename, drive, dir, fname, ext);
        mbcs_makepath( filename, drive, dir, fname, e);
    }
    return (filename);
}

const char *S9xGetFilenameInc (const char *e)
{
    static char filename [_MAX_PATH + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fname [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];
    char *ptr;

    if (strlen (GUI.FreezeFileDir))
    {
        mbcs_splitpath (Memory.ROMFilename, drive, dir, fname, ext);
        strcpy (filename, GUI.FreezeFileDir);
        strcat (filename, TEXT("\\"));
        strcat (filename, fname);
        ptr = filename + strlen (filename);
        strcat (filename, TEXT("00/"));
        strcat (filename, e);
    }
    else
    {
        mbcs_splitpath (Memory.ROMFilename, drive, dir, fname, ext);
        strcat (fname, TEXT("00/"));
        mbcs_makepath (filename, drive, dir, fname, e);
        ptr = strstr (filename, TEXT("00/"));
    }

    do
    {
        if (++*(ptr + 2) > '9')
        {
            *(ptr + 2) = '0';
            if (++*(ptr + 1) > '9')
            {
                *(ptr + 1) = '0';
                if (++*ptr > '9')
                    break;
            }
        }
    } while (_access (filename, 0) == 0);

    return (filename);
}

static int spcnum = 0;

const char *S9xGetSPCFilename ()
{
    static char filename [_MAX_PATH + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fname [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];


    if (strlen (Memory.SPCDumpDirectory))
    {
        mbcs_splitpath (Memory.ROMFilename, drive, dir, fname, ext);
        strcpy (filename, Memory.SPCDumpDirectory);
        strcat (filename, "\\");
        strcat (filename, fname);
		strcat (filename, "_000");
        strcat (filename, ".spc");
		mbcs_splitpath (filename, drive, dir, fname, ext);
    }
    else
    {
        mbcs_splitpath (Memory.ROMFilename, drive, dir, fname, ext);
		strcat (fname, "_000");
        mbcs_makepath( filename, drive, dir, fname, ".spc");
		mbcs_splitpath (filename, drive, dir, fname, ext);
    }
	spcnum = 0;
	while((int)GetFileAttributes(filename) != -1 && spcnum != 999)
	{
		++spcnum;
		wsprintf(&fname[lstrlen(fname) - 3], "%03d", spcnum);
		mbcs_makepath( filename, drive, dir, fname, ext);
	}
    return (const char *)filename;
}


static char SavedSPCNumberAndExtStr[10];

char *S9xwGetSavedSPCNumberAndExtStr()
{
	SavedSPCNumberAndExtStr[0] = '_';
	wsprintf(&SavedSPCNumberAndExtStr[1], "%03d", spcnum);
	lstrcat(SavedSPCNumberAndExtStr, ".spc");
	return SavedSPCNumberAndExtStr;
}

static char imgext [_MAX_EXT] = ".bmp";
static int imgnum = 0;

char *S9xwGetScreenImageFilename (char *ImageTypeExt)
{
	static char filename [_MAX_PATH + 1];
	char drive [_MAX_DRIVE + 1];
	char dir [_MAX_DIR + 1];
	char fname [_MAX_FNAME + 1];

	mbcs_splitpath (Memory.ROMFilename, drive, dir, fname, imgext);
	lstrcpy(imgext, ImageTypeExt);
	CharLower(imgext);

	if (strlen (Memory.ScreenShotDirectory))
	{
		strcpy (filename, Memory.ScreenShotDirectory);
		strcat (filename, "\\");
		strcat (filename, fname);
		strcat (filename, "_000");
		strcat (filename, imgext);
		mbcs_splitpath (filename, drive, dir, fname, imgext);
	}
	else
	{
		strcat (fname, "_000");
		mbcs_makepath( filename, drive, dir, fname, imgext);
		mbcs_splitpath (filename, drive, dir, fname, imgext);
	}
	
	imgnum = 0;
	while((int)GetFileAttributes(filename) != -1 && imgnum != 999)
	{
		++imgnum;
		wsprintf(&fname[lstrlen(fname) - 3], "%03d", imgnum);
		mbcs_makepath( filename, drive, dir, fname, imgext);
	}

	return (filename);
}

extern "C" const char *S9xGetNumFilename (const char *dirstr, const char *fext, int *numsaved)
{
	static char filename [_MAX_PATH + 1];
	char drive [_MAX_DRIVE + 1];
	char dir [_MAX_DIR + 1];
	char fname [_MAX_FNAME + 1];
	char ext [_MAX_EXT + 1];
	int num = 0;

	mbcs_splitpath (Memory.ROMFilename, drive, dir, fname, ext);
	lstrcpy(ext, fext);
	CharLower(ext);

	if (lstrlen (dirstr))
	{
		if((lstrlen(dirstr) + 1 + lstrlen(fname) + 4 + lstrlen(ext)) > _MAX_PATH) {
			if(numsaved)
				*numsaved = num;
			return NULL;
		}

		lstrcpy (filename, dirstr);
		lstrcat (filename, "\\");
		lstrcat (filename, fname);
		lstrcat (filename, "_000");
		lstrcat (filename, ext);
		mbcs_splitpath (filename, drive, dir, fname, ext);
	}
	else
	{
		if((lstrlen(drive) + lstrlen(dir) + lstrlen(fname) + 4 + lstrlen(ext)) > _MAX_PATH) {
			if(numsaved)
				*numsaved = num;
			return NULL;
		}

		lstrcat (fname, "_000");
		mbcs_makepath( filename, drive, dir, fname, ext);
		mbcs_splitpath (filename, drive, dir, fname, ext);
	}
	
	while(GetFileAttributes(filename) != 0xffffffff && num != 999)
	{
		++num;
		wsprintf(&fname[lstrlen(fname) - 3], "%03d", num);
		mbcs_makepath( filename, drive, dir, fname, ext);
	}

	if(numsaved)
		*numsaved = num;

	return (const char *)filename;
}

static char SavedScreenImageNumberAndTypeStr[10];

char *S9xwGetSavedScreenImageNumberAndTypeStr()
{
	SavedScreenImageNumberAndTypeStr[0] = '_';
	wsprintf(&SavedScreenImageNumberAndTypeStr[1], "%03d", imgnum);
	lstrcat(SavedScreenImageNumberAndTypeStr, imgext);
	return SavedScreenImageNumberAndTypeStr;
}

bool8 S9xOpenSnapshotFile(const char *fname, bool8 read_only, STREAM *file)
{
    char filename [_MAX_PATH + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fn [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];

    mbcs_splitpath( fname, drive, dir, fn, ext);
    mbcs_makepath( filename, drive, dir, fn, ext[0] == '\0' ? ".000" : ext);

    if (read_only)
    {
	if ((*file = OPEN_STREAM (filename, "rb")))
	    return (TRUE);
    }
    else
    {
	if ((*file = OPEN_STREAM (filename, "wb")))
	    return (TRUE);
        FILE *fs = fopen (filename, "rb");
        if (fs)
        {
            sprintf (String, "Freeze file \"%s\" exists but is read only",
                     filename);
            fclose (fs);
            S9xMessage (S9X_ERROR, S9X_FREEZE_FILE_NOT_FOUND, String);
        }
        else
        {
            sprintf (String, "Cannot create freeze file \"%s\". Directory is read-only or does not exist.", filename);
            
            S9xMessage (S9X_ERROR, S9X_FREEZE_FILE_NOT_FOUND, String);
        }
    }
    return (FALSE);
}

void S9xCloseSnapshotFile( STREAM file)
{
    CLOSE_STREAM (file);
}

void S9xMessage (int, int, const char *str)
{
#ifdef DEBUGGER
  //static FILE *out = NULL;

  //if (out == NULL)
  //        out = fopen ("out.txt", "w");

  //    fprintf (out, "%s\n", str);
#endif

    S9xSetInfoString (str);
}

bool8 S9xOpenSoundDevice (int mode, bool8 pStereo, int BufferSize)
{
    return (TRUE);
}

int superscope_pause = 0;
int superscope_turbo = 0;
static RECT srcRect = {0, 512, 0, 448};
static RECT dstRect = {0, 512, 0, 448};

bool8 S9xReadSuperScopePosition (int &x, int &y, uint32 &buttons)
{
    x = (int) ((GUI.MouseX - dstRect.left) * (256.0 / (dstRect.right - dstRect.left)));
    //not all games are 224
	y = (int) ((GUI.MouseY - dstRect.top) * ((double)PPU.ScreenHeight / (dstRect.bottom - dstRect.top)));
    buttons = GUI.MouseButtons | (superscope_turbo << 2) | 
              (superscope_pause << 3);

	superscope_pause=FALSE;

    return (TRUE);
}

void S9xSyncSpeed( void)
{
    if ((!Settings.TurboMode && Settings.SkipFrames == AUTO_FRAMERATE) ||
		(XxxMovie.Replay_rec == TRUE || XxxMovie.Replay_play == TRUE))
    {
        LONG prev;
        BOOL success;
        static unsigned long XXX_FrameTime;

        if ((success = ReleaseSemaphore (GUI.FrameTimerSemaphore, 1, &prev)) &&
            prev == 0)
        {
            // Mop up the ReleaseSemaphore above...
            WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
            // ...and then wait for the real sync signal
            WaitForSingleObject (GUI.FrameTimerSemaphore, 100);
            IPPU.RenderThisFrame = TRUE;
            IPPU.SkippedFrames = 0;
            XXX_FrameTime = 0;
        }
        else
        {
          if (success)
              WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
          WaitForSingleObject (GUI.FrameTimerSemaphore, 0);

	    if (IPPU.SkippedFrames < Settings.AutoMaxSkipFrames && prev > 1)
	    {
		IPPU.SkippedFrames++;
		IPPU.RenderThisFrame = FALSE;
		XXX_FrameTime = 0;
	    }
	    else
	    {

	                if (IPPU.SkippedFrames==0 && XXX_FrameTime > 3)
	                {
		            IPPU.SkippedFrames++;
		            IPPU.RenderThisFrame = FALSE;
		            XXX_FrameTime = 0;
                        }
                        else
                        {
		            IPPU.RenderThisFrame = TRUE;
		            IPPU.SkippedFrames = 0;
		            XXX_FrameTime++;
                        }
	    }
        }
    }
    else
    {
	if (++IPPU.FrameSkip >= (Settings.TurboMode ? Settings.TurboSkipFrames
                                                    : Settings.SkipFrames))
	{
	    IPPU.FrameSkip = 0;
	    IPPU.SkippedFrames = 0;
	    IPPU.RenderThisFrame = TRUE;
	}
	else
	{
	    IPPU.SkippedFrames++;
	    IPPU.RenderThisFrame = FALSE;
	}
    }
}

const char *S9xBasename (const char *f)
{
	const char *p;
	const char *pp = f;
	size_t slen = lstrlen(f);
	//SJIS string
	for (p = f; p < (f + slen); p++) {
		if(*(p + 1) && ((*(unsigned char *)p >= 0x81 && *(unsigned char *)p <= 0x9f) ||
						(*(unsigned char *)p >= 0xe0 && *(unsigned char *)p <= 0xfc)) &&
		   ((*(unsigned char *)(p + 1) >= 0x40 && *(unsigned char *)(p + 1) <= 0x7e) ||
			(*(unsigned char *)(p + 1) >= 0x80 && *(unsigned char *)(p + 1) <= 0xfc)))
			++p;
		else if (*p == (const char)'\\')
			pp = (p + 1);
	}
	return pp;
}

bool8 S9xReadMousePosition (int which, int &x, int &y, uint32 &buttons)
{
    if (which == 0)
    {
        x = GUI.MouseX;
        y = GUI.MouseY;
        buttons = GUI.MouseButtons;
        return (TRUE);
    }

    return (FALSE);
}

PALETTEENTRY S9xPaletteEntry[256];
void S9xSetPalette( void)
{
    LPDIRECTDRAWPALETTE lpDDTemp;
    uint16 Brightness = IPPU.MaxBrightness * 140;

    // Only update the palette structures if needed
    if( GUI.ScreenDepth == 8)
    {
        if (Settings.SixteenBit)
        {
            for (int i = 0; i < 256; i++)
            {
                S9xPaletteEntry[i].peRed   = FixedColours [i].red;
                S9xPaletteEntry[i].peGreen = FixedColours [i].green;
                S9xPaletteEntry[i].peBlue  = FixedColours [i].blue;
            }
        }
        else
        {
            for (int i = 0; i < 256; i ++)
            {	
                S9xPaletteEntry[i].peRed   = (((PPU.CGDATA [i] >>  0) & 0x1F) * Brightness) >> 8;
                S9xPaletteEntry[i].peGreen = (((PPU.CGDATA [i] >>  5) & 0x1F) * Brightness) >> 8;
                S9xPaletteEntry[i].peBlue  = (((PPU.CGDATA [i] >> 10) & 0x1F) * Brightness) >> 8;
            }
        }
		DirectX.lpDDSPrimary2->GetPalette(&lpDDTemp);
		if(lpDDTemp != DirectX.lpDDPalette)
			DirectX.lpDDSPrimary2->SetPalette(DirectX.lpDDPalette);
        DirectX.lpDDPalette->SetEntries (0, 0, 256, S9xPaletteEntry);
    }
}

static BYTE *ScreenBuffer = NULL;
static BYTE *SubScreenBuffer = NULL;
static BYTE *ZBuffer = NULL;
static BYTE *SubZBuffer = NULL;

static void FreeScreenBuffer()
{
	if(ScreenBuffer) {
		GlobalFree((HGLOBAL)ScreenBuffer);
		ScreenBuffer = NULL;
	}
	if(SubScreenBuffer) {
		GlobalFree((HGLOBAL)SubScreenBuffer);
		SubScreenBuffer = NULL;
	}
	if(ZBuffer) {
		GlobalFree((HGLOBAL)ZBuffer);
		ZBuffer = NULL;
	}
	if(SubZBuffer) {
		GlobalFree((HGLOBAL)SubZBuffer);
		SubZBuffer = NULL;
	}
	return;
}

static bool8 AllocScreenBuffer()
{
	ScreenBuffer = (BYTE *)GlobalAlloc(GPTR, 512*478*2);
	SubScreenBuffer = (BYTE *)GlobalAlloc(GPTR, 512*478*2);
	ZBuffer = (BYTE *)GlobalAlloc(GPTR, 512*478);
	SubZBuffer = (BYTE *)GlobalAlloc(GPTR, 512*478);

	if(!ScreenBuffer || !SubScreenBuffer || !ZBuffer || !SubZBuffer) {
		FreeScreenBuffer();
		return FALSE;
	}
	return TRUE;
}

//static bool8 locked_surface = FALSE;
bool8 locked_surface = FALSE;
static SSurface Src, Dst;

static bool8 force_S9xDeinitUpdate = FALSE;

bool8 S9xInitUpdate( void)
{
    GFX.SubScreen = SubScreenBuffer;
    GFX.ZBuffer = ZBuffer;
    GFX.SubZBuffer = SubZBuffer;

    if (force_S9xDeinitUpdate == FALSE &&
		(GUI.Scale <= 1 ||
		 (DirectX.DisableHardwareScale == FALSE &&
		  ((GUI.Scale == 3 && IPPU.Interlace) ||
		   (GUI.Scale <= 7 && GUI.Scale >= 4 && (PPU.BGMode == 5 || PPU.BGMode == 6 || IPPU.Interlace)) ||
		   (GUI.Scale >= 8 && IPPU.Interlace)))) && 
		!GUI.NeedDepthConvert &&
        (DirectX.lpDDSOffScreen2 && LockSurface2 (DirectX.lpDDSOffScreen2, &Src))) {

        locked_surface = TRUE;

        GFX.RealPitch = GFX.Pitch2 = GFX.Pitch = Src.Pitch;
        GFX.Screen = Src.Surface;
	    DirectX.lpDDSOffScreen2->Unlock (Src.Surface);

		GFX.ZPitch = GFX.Pitch;
		if (Settings.SixteenBit)
			GFX.ZPitch >>= 1;
    }
    else
    {
        locked_surface = FALSE;
//        GFX.RealPitch = GFX.Pitch = 512 * 2;
//        GFX.Screen = ScreenBuffer;
		if (Settings.SixteenBit)
		{
			GFX.RealPitch = GFX.Pitch2 = GFX.Pitch = 512 * 2;
			GFX.ZPitch = GFX.Pitch;
			GFX.ZPitch >>= 1;
			GFX.Screen = ScreenBuffer;
		}
		else
		{
			GFX.RealPitch = GFX.Pitch2 = GFX.Pitch = 512;
			GFX.ZPitch = GFX.Pitch;
			GFX.Screen = ScreenBuffer;
			GFX.SubScreen = &ScreenBuffer[512*478];
		}
    }
    GFX.Delta = (GFX.SubScreen - GFX.Screen) >> 1;

    if (Settings.SixteenBit)
    {
        GFX.PPL = GFX.Pitch >> 1;
        GFX.PPLx2 = GFX.Pitch;
    }
    else
    {
        GFX.PPL = GFX.Pitch;
        GFX.PPLx2 = GFX.Pitch * 2;
    }
//    GFX.Pitchx2 = GFX.Pitch * 2;
    return (TRUE);
}

#define IS_DONT_RENDER()\
	((force_S9xDeinitUpdate == FALSE && GUI.hWnd != GetForegroundWindow()) || \
	 (!DirectX.Windowed && ExtendHeight &&								\
	  (GUI.CurFullscreenDisplayModeIndex <= 1 ||							\
	 (GUI.Scale && (GUI.CurFullscreenDisplayModeIndex == 12 || GUI.CurFullscreenDisplayModeIndex == 13)))))

#define CLEAR_RENDER_SCREEN()\
	if(/*OldRenderMethod != RenderMethod || */ExtendHeight_previos != ExtendHeight) { \
    if(DirectX.Windowed) {\
	  RECT clrRect;\
	  HDC clrhdc;\
	  clrRect = GUI.ClientRect;\
	  clrhdc = GetDC(GUI.hWnd);\
	  FillRect(clrhdc, &clrRect, (HBRUSH)GetStockObject(BLACK_BRUSH));\
	  ReleaseDC(GUI.hWnd, clrhdc);\
	}\
	else {\
	  DirectX_ClearPrimarySurface();\
	  if(!DirectX.Windowed && DirectX.DoubleBuffered)\
		DirectX_ClearBackbufferSurface();\
	}\
  }

#define RENDERED_EXIT(r)\
	GUI.SizeHistory [GUI.FlipCounter % DirectX.NumFlipFrames] = dstRect;\
	GUI.FlipCounter++;\
	return (r);

bool8 S9xDeinitUpdate(int Width, int Height, bool8 sixteen_bit)
{
    LPDIRECTDRAWSURFACE2 lpDDSurface2 = NULL;
    bool PrimarySurfaceLockFailed = FALSE;
	static int update_method = 0;

	DisplayStatusInfo();

	if(IS_DONT_RENDER()) {
		RENDERED_EXIT(TRUE);
	}

	if(Options.DisplayScreenBufferAll && GUI.Scale > 0) {
		Width = 512;
		Height = 478;
	}
    Src.Width = Width;
    Src.Height = Height;
    Src.Pitch = GFX.RealPitch;
    Src.Surface = GFX.Screen;

	DoubleHeight = Src.Height > 240;
	DoubleWidth  = Src.Width == 512;
	ExtendHeight = Src.Height == 239 || Src.Height == 478;

	srcRect.top    = 0;
	srcRect.bottom = Height;
	srcRect.left   = 0;
	srcRect.right  = Width;

	dstRect = GUI.ClientRect;
	bool8 dstrect_ok = (dstRect.right - dstRect.left) >= (GUI.Scale > 0 ? 512 : 256) &&
		(dstRect.bottom - dstRect.top) >= (GUI.Scale > 0 ? (ExtendHeight ? 478 : 448) : (ExtendHeight ? 239 : 224));

	// OffScreen has beed rendered image
	// It not needs render
	//
	// GUI.Scale <= 1 && !GUI.NeedDepthConvert
	//
    if (locked_surface &&
		(GUI.dRendering == FALSE || GUI.Stretch) &&
		(GUI.Scale <= 1 ||
		 (DirectX.DisableHardwareScale == FALSE &&
		  GUI.Scale >= 3))) {

		update_method = 0;
		ResetRenderMethod();
		CLEAR_RENDER_SCREEN();

	    POINT p;
		// this is OK on fullscreen
		// dstRect = GUI.ClientRect;
		// stretch
		if (GUI.Stretch == 1) {
			p.x = p.y = 0;
			ClientToScreen (GUI.hWnd, &p);
			dstRect.right = (dstRect.right - dstRect.left) + p.x;
			dstRect.bottom = (dstRect.bottom - dstRect.top) + p.y;
			dstRect.top = p.y;
			dstRect.left = p.x;
		}
		else if (GUI.Stretch == 2) {
			int dst_width = srcRect.right - srcRect.left;
			int dst_height = srcRect.bottom - srcRect.top;

			if (GUI.Scale > 0) {
			    // no Wx2 hires
			    if (dst_width < 257)
				    dst_width *= 2;
			    // no Hx2 hires
				if (dst_height < 240)
				    dst_height *= 2;
			}

			if((dstRect.right - dstRect.left) >= dst_width &&
				(dstRect.bottom - dstRect.top) >= (dst_height * ((dstRect.right - dstRect.left) / dst_width))) {
				dst_width = 256;
				dst_height = ExtendHeight ? 239 : 224;
				long scalemulw = ((dstRect.right - dstRect.left) / dst_width);
				long scalemulh = ((dstRect.bottom - dstRect.top) / dst_height);
				long scalemul = scalemulw < scalemulh ? scalemulw : scalemulh;
				dst_width *= scalemul;
				p.x = ((dstRect.right - dstRect.left) - dst_width) / 2;
				dst_height *= scalemul;
				p.y = ((dstRect.bottom - dstRect.top) - dst_height) / 2;
			}
			else if((dstRect.bottom - dstRect.top) >= dst_height &&
				(dstRect.right - dstRect.left) >= (dst_width * ((dstRect.bottom - dstRect.top) / dst_height))) {
				dst_width = 256;
				dst_height = ExtendHeight ? 239 : 224;
				long scalemulw = ((dstRect.right - dstRect.left) / dst_width);
				long scalemulh = ((dstRect.bottom - dstRect.top) / dst_height);
				long scalemul = scalemulw < scalemulh ? scalemulw : scalemulh;
				dst_width *= scalemul;
				p.x = ((dstRect.right - dstRect.left) - dst_width) / 2;
				dst_height *= scalemul;
				p.y = ((dstRect.bottom - dstRect.top) - dst_height) / 2;
			}
			else {
				if((dstRect.right - dstRect.left) < dst_width) {
					p.x = 0;
					dst_width = dstRect.right - dstRect.left;
				}
				else {
					p.x = ((dstRect.right - dstRect.left) - dst_width) / 2;
				}
				if((dstRect.bottom - dstRect.top) < dst_height) {
					p.y = 0;
					dst_height = dstRect.bottom - dstRect.top;
				}
				else {
					p.y = ((dstRect.bottom - dstRect.top) - dst_height) / 2;
				}
			}

			ClientToScreen (GUI.hWnd, &p);
			dstRect.top = p.y;
			dstRect.left = p.x;
			dstRect.bottom = dstRect.top + dst_height;
			dstRect.right  = dstRect.left + dst_width;
		}
		else {
			int dst_width = srcRect.right - srcRect.left;
			int dst_height = srcRect.bottom - srcRect.top;

			if (GUI.Scale > 0) {
			    // no Wx2 hires
			    if (dst_width < 257)
				    dst_width *= 2;
			    // no Hx2 hires
				if (dst_height < 240)
				    dst_height *= 2;
			}
			p.x = ((dstRect.right - dstRect.left) - dst_width) / 2;
			p.y = ((dstRect.bottom - dstRect.top) - dst_height) / 2;
			ClientToScreen (GUI.hWnd, &p);

			dstRect.top = p.y;
			dstRect.left = p.x;
			dstRect.bottom = dstRect.top + dst_height;
			dstRect.right  = dstRect.left + dst_width;
		}

		lpDDSurface2 = DirectX.lpDDSOffScreen2;
		lpDDSurface2->Unlock(Src.Surface);	

		bool8 _backbuffer;

		//fullscreen DoubleBuffered
		if(!DirectX.Windowed && DirectX.DoubleBuffered) {
		    _backbuffer = TRUE;
		    lpDDSurface2 = DirectX.lpDDSBack2;
		}
		else {
		    _backbuffer = FALSE;
		    lpDDSurface2 = DirectX.lpDDSPrimary2;
		}

		if(DirectX.wait_for_vsync && _backbuffer == FALSE)
			DirectX.lpDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, 0);

		if (lpDDSurface2->Blt (&dstRect, DirectX.lpDDSOffScreen2, &srcRect, DDBLT_WAIT, NULL) != DD_OK) {
			RENDERED_EXIT(FALSE);
		}

		if(_backbuffer)
		    DirectX.lpDDSPrimary2->Flip(NULL, DDFLIP_WAIT);
    }
	//Directly Render to screen
	else if(force_S9xDeinitUpdate == FALSE &&
			dstrect_ok && !DirectX.lpDDClipper && GUI.dRendering && GUI.Stretch == 0) {

		if(update_method != 1) {
			update_method = 1;
			ResetRenderMethod();
		}

		int dst_width = srcRect.right - srcRect.left;
		int dst_height = srcRect.bottom - srcRect.top;

		if (GUI.Scale > 0) {
		    // no hires
			if (dst_width <= 256)
				dst_width *= 2;
			// no x2 height (??)
			if (dst_height < 240)
				dst_height *= 2;
		}

		POINT p;
		// dstRect = GUI.ClientRect;
		p.x = ((dstRect.right - dstRect.left) - dst_width) / 2;
		p.y = ((dstRect.bottom - dstRect.top) - dst_height) / 2;
		ClientToScreen (GUI.hWnd, &p);

		dstRect.top = p.y;
		dstRect.left = p.x;
		dstRect.bottom = dstRect.top + dst_height;
		dstRect.right  = dstRect.left + dst_width;

		bool8 _backbuffer;

		//fullscreen DoubleBuffered
		if(!DirectX.Windowed && DirectX.DoubleBuffered) {
		    _backbuffer = TRUE;
		    lpDDSurface2 = DirectX.lpDDSBack2;
			DirectX.NumFlipFrames = 2;
		}
		else {
		    _backbuffer = FALSE;
		    lpDDSurface2 = DirectX.lpDDSPrimary2;
			DirectX.NumFlipFrames = 1;
		}

		if (LockSurface2 (lpDDSurface2, &Dst) == FALSE) {
			ResetRenderMethod();
			RENDERED_EXIT(FALSE);
		}

		SelectRenderMethod(TRUE);
		CLEAR_RENDER_SCREEN();

		if(DirectX.wait_for_vsync && _backbuffer == FALSE)
			DirectX.lpDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, 0);

		RenderMethod(&Src, &Dst, &srcRect, &dstRect);
		lpDDSurface2->Unlock(Dst.Surface);
		if(_backbuffer)
		    DirectX.lpDDSPrimary2->Flip(NULL, DDFLIP_WAIT);
	}
	// render offscreen & flip or blit
	else {

		if(update_method != 2) {
			update_method = 2;
			ResetRenderMethod();
		}

		int dst_width = srcRect.right - srcRect.left;
		int dst_height = srcRect.bottom - srcRect.top;

		bool8 _backbuffer;

		if(!DirectX.Windowed && DirectX.DoubleBuffered) {
		    _backbuffer = TRUE;
		}
		else {
		    _backbuffer = FALSE;
		}

		bool8 softwarescale;

		// hardware scale ?
		if(GUI.Stretch ||
		   GUI.Scale == 1 ||
		   (DirectX.DisableHardwareScale == FALSE &&
			(GUI.Scale == 3 ||
			 (GUI.Scale <= 7 && GUI.Scale >= 4 && (DoubleWidth ^ DoubleHeight)) ||
			 (GUI.Scale >= 8 && DoubleHeight == TRUE)))) {

			softwarescale = FALSE;

		    lpDDSurface2 = DirectX.lpDDSOffScreen2;
			DirectX.NumFlipFrames = 1;

			if (GUI.Scale == 2) {
			    // no hires
			    if (dst_width <= 256) {
				    dst_width *= 2;
				}
				// no x2 height (??)
				if (dst_height < 240) {
				    dst_height *= 2;
				}
			}
			else if (GUI.Scale == 3 && DoubleHeight == FALSE) {
			    dst_height *= 2;
			}
			else if (GUI.Scale <= 7 && GUI.Scale >= 4 &&
					 DoubleWidth == FALSE && DoubleHeight == FALSE) {
				dst_width *= 2;
				dst_height *= 2;
			}
			else if (GUI.Scale >= 8 && DoubleHeight == FALSE) {
			    // no hires
			    if (dst_width <= 256) {
				    dst_width *= 2;
				}
			    dst_height *= 2;
			}
		}
		// No hardware scale
		else {
			softwarescale = TRUE;

		    if(_backbuffer) {
			    lpDDSurface2 = DirectX.lpDDSBack2;
				DirectX.NumFlipFrames = 2;
			}
			else {
			    lpDDSurface2 = DirectX.lpDDSOffScreen2;
				DirectX.NumFlipFrames = 1;
			}
			if (GUI.Scale > 0) {
			    // no hires
			    if (dst_width <= 256) {
				    dst_width *= 2;
				}
				// no x2 height (??)
				if (dst_height < 240) {
				    dst_height *= 2;
				}
			}
		}

		if (LockSurface2 (lpDDSurface2, &Dst) == FALSE) {
			ResetRenderMethod();
			RENDERED_EXIT(FALSE);
		}

		dstRect.left = (Dst.Width - dst_width) / 2;
		dstRect.right = dstRect.left + dst_width;
		dstRect.top = (Dst.Height - dst_height) / 2;
		dstRect.bottom = dstRect.top + dst_height;

		SelectRenderMethod(softwarescale);
		CLEAR_RENDER_SCREEN();
		RenderMethod(&Src, &Dst, &srcRect, &dstRect);

		lpDDSurface2->Unlock(Dst.Surface);

		// DoubleBuffer
		if(_backbuffer && lpDDSurface2 == DirectX.lpDDSBack2) {
		    DirectX.lpDDSPrimary2->Flip(NULL, DDFLIP_WAIT);
			RENDERED_EXIT(TRUE);
		}

		// OffScreen -> Primary
		Src = Dst;
		srcRect = dstRect;

	    POINT p;
		// this is OK on fullscreen
		dstRect = GUI.ClientRect;
		// stretch
		if (GUI.Stretch == 1) {
			p.x = p.y = 0;
			ClientToScreen (GUI.hWnd, &p);
			dstRect.right = (dstRect.right - dstRect.left) + p.x;
			dstRect.bottom = (dstRect.bottom - dstRect.top) + p.y;
			dstRect.top = p.y;
			dstRect.left = p.x;
		}
		else if (GUI.Stretch == 2) {
			int dst_width = srcRect.right - srcRect.left;
			int dst_height = srcRect.bottom - srcRect.top;

			if (GUI.Scale > 0) {
			    // no Wx2 hires
			    if (dst_width < 257)
				    dst_width *= 2;
			    // no Hx2 hires
				if (dst_height < 240)
				    dst_height *= 2;
			}

			if((dstRect.right - dstRect.left) >= dst_width &&
				(dstRect.bottom - dstRect.top) >= (dst_height * ((dstRect.right - dstRect.left) / dst_width))) {
				dst_width = 256;
				dst_height = ExtendHeight ? 239 : 224;
				long scalemulw = ((dstRect.right - dstRect.left) / dst_width);
				long scalemulh = ((dstRect.bottom - dstRect.top) / dst_height);
				long scalemul = scalemulw < scalemulh ? scalemulw : scalemulh;
				dst_width *= scalemul;
				p.x = ((dstRect.right - dstRect.left) - dst_width) / 2;
				dst_height *= scalemul;
				p.y = ((dstRect.bottom - dstRect.top) - dst_height) / 2;
			}
			else if((dstRect.bottom - dstRect.top) >= dst_height &&
				(dstRect.right - dstRect.left) >= (dst_width * ((dstRect.bottom - dstRect.top) / dst_height))) {
				dst_width = 256;
				dst_height = ExtendHeight ? 239 : 224;
				long scalemulw = ((dstRect.right - dstRect.left) / dst_width);
				long scalemulh = ((dstRect.bottom - dstRect.top) / dst_height);
				long scalemul = scalemulw < scalemulh ? scalemulw : scalemulh;
				dst_width *= scalemul;
				p.x = ((dstRect.right - dstRect.left) - dst_width) / 2;
				dst_height *= scalemul;
				p.y = ((dstRect.bottom - dstRect.top) - dst_height) / 2;
			}
			else {
				if((dstRect.right - dstRect.left) < dst_width) {
					p.x = 0;
					dst_width = dstRect.right - dstRect.left;
				}
				else {
					p.x = ((dstRect.right - dstRect.left) - dst_width) / 2;
				}
				if((dstRect.bottom - dstRect.top) < dst_height) {
					p.y = 0;
					dst_height = dstRect.bottom - dstRect.top;
				}
				else {
					p.y = ((dstRect.bottom - dstRect.top) - dst_height) / 2;
				}
			}

			ClientToScreen (GUI.hWnd, &p);
			dstRect.top = p.y;
			dstRect.left = p.x;
			dstRect.bottom = dstRect.top + dst_height;
			dstRect.right  = dstRect.left + dst_width;
		}
		else {
			dst_width = srcRect.right - srcRect.left;
			dst_height = srcRect.bottom - srcRect.top;

			if (GUI.Scale > 0) {
			    // no Wx2 hires
			    if (dst_width < 257)
				    dst_width *= 2;
			    // no Hx2 hires
				if (dst_height < 240)
				    dst_height *= 2;
			}
			p.x = ((dstRect.right - dstRect.left) - dst_width) / 2;
			p.y = ((dstRect.bottom - dstRect.top) - dst_height) / 2;
			ClientToScreen (GUI.hWnd, &p);

			dstRect.top = p.y;
			dstRect.left = p.x;
			dstRect.bottom = dstRect.top + dst_height;
			dstRect.right  = dstRect.left + dst_width;
		}

		//fullscreen DoubleBuffered
		if(_backbuffer) {
		    lpDDSurface2 = DirectX.lpDDSBack2;
		}
		else {
		    lpDDSurface2 = DirectX.lpDDSPrimary2;
		}

		if(DirectX.wait_for_vsync && _backbuffer == FALSE)
			DirectX.lpDD->WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, 0);

		if (lpDDSurface2->Blt (&dstRect, DirectX.lpDDSOffScreen2, &srcRect, DDBLT_WAIT, NULL) != DD_OK) {
			RENDERED_EXIT(FALSE);
		}

		if(_backbuffer)
		    DirectX.lpDDSPrimary2->Flip(NULL, DDFLIP_WAIT);
	}	
	RENDERED_EXIT(TRUE);
	// END
}

void ScreenRedraw(HWND hWnd)
{
	if(!DirectX.DDAvailable || !DirectX.Windowed)
		return;

	RefreshRectParams();

	PAINTSTRUCT ps;
	HDC hdc = BeginPaint(hWnd, &ps);
	EndPaint(hWnd, &ps);

	unsigned char GUI_dRendering_bak = GUI.dRendering;
	GUI.dRendering = FALSE;
	force_S9xDeinitUpdate = TRUE;

	// screenbuffer == OffScreen
	if(locked_surface) {
		locked_surface = FALSE;
		uint8 *sb = ScreenBuffer;
		for(int y = 0; y < IPPU.RenderedScreenHeight; y++)
			memcpy(sb + y * (512 * (Settings.SixteenBit ? 2 : 1)), GFX.Screen + y * GFX.Pitch,
				   IPPU.RenderedScreenWidth * (Settings.SixteenBit ? 2 : 1));
	}
	S9xInitUpdate();

	int width = 256 * (DoubleWidth ? 2 : 1);
	int height = (ExtendHeight ? 239 : 224) * (DoubleHeight ? 2 : 1);

	S9xDeinitUpdate(width, height, Settings.SixteenBit);
	force_S9xDeinitUpdate = FALSE;
	GUI.dRendering = GUI_dRendering_bak;
	return;
}

int ffs (uint32 mask)
{
    int m = 0;
    if (mask)
    {
        while (!(mask & (1 << m)))
            m++;

        return (m);
    }

    return (0);
}

void S9xSetWinPixelFormat ()
{
    extern int Init_2xSaI (uint32 BitFormat);
	extern void hq2x_init(unsigned);

    S9xSetRenderPixelFormat (RGB565);
    Init_2xSaI (565);
	hq2x_init(16);
    GUI.NeedDepthConvert = FALSE;

#if 0
    if (VOODOO_MODE)
    {
        GUI.ScreenDepth = 16;
        GUI.RedShift = 0;
        GUI.GreenShift = 5;
        GUI.BlueShift = 11;
        Settings.SixteenBit = TRUE;
    }
    else
    if (OPENGL_MODE)
    {
        GUI.ScreenDepth = 16;
        GUI.RedShift = 0;
        GUI.GreenShift = 5;
        GUI.BlueShift = 11;
        Settings.SixteenBit = TRUE;
    }
    else
    {
#endif
        GUI.ScreenDepth = DirectX.DDPixelFormat.dwRGBBitCount;
        if (GUI.ScreenDepth == 15)
            GUI.ScreenDepth = 16;

        GUI.RedShift = ffs (DirectX.DDPixelFormat.dwRBitMask);
        GUI.GreenShift = ffs (DirectX.DDPixelFormat.dwGBitMask);
        GUI.BlueShift = ffs (DirectX.DDPixelFormat.dwBBitMask);

        if((DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
           GUI.ScreenDepth == 16 &&
           DirectX.DDPixelFormat.dwRBitMask == 0xF800 &&
           DirectX.DDPixelFormat.dwGBitMask == 0x07E0 &&
           DirectX.DDPixelFormat.dwBBitMask == 0x001F)
        {
			GUI.GreenShift += 1;
            S9xSetRenderPixelFormat (RGB565);
            Init_2xSaI (565);
			hq2x_init(16);
        }
        else
            if( (DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
                GUI.ScreenDepth == 16 &&
                DirectX.DDPixelFormat.dwRBitMask == 0x7C00 &&
                DirectX.DDPixelFormat.dwGBitMask == 0x03E0 &&
                DirectX.DDPixelFormat.dwBBitMask == 0x001F)
            {
                S9xSetRenderPixelFormat (RGB555);
                Init_2xSaI (555);
				hq2x_init(15);
            }
            else
                if((DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
                   GUI.ScreenDepth == 16 &&
                   DirectX.DDPixelFormat.dwRBitMask == 0x001F &&
                   DirectX.DDPixelFormat.dwGBitMask == 0x07E0 &&
                   DirectX.DDPixelFormat.dwBBitMask == 0xF800)
                {
					GUI.GreenShift += 1;
                    S9xSetRenderPixelFormat (BGR565);
                    Init_2xSaI (565);
					hq2x_init(16);
                }
                else
                    if( (DirectX.DDPixelFormat.dwFlags&DDPF_RGB) != 0 &&
                        GUI.ScreenDepth == 16 &&
                        DirectX.DDPixelFormat.dwRBitMask == 0x001F &&
                        DirectX.DDPixelFormat.dwGBitMask == 0x03E0 &&
                        DirectX.DDPixelFormat.dwBBitMask == 0x7C00)
                    {
                        S9xSetRenderPixelFormat (BGR555);
                        Init_2xSaI (555);
						hq2x_init(15);
                    }
                    else
                        if (DirectX.DDPixelFormat.dwRGBBitCount == 8 ||
                            DirectX.DDPixelFormat.dwRGBBitCount == 24 ||
                            DirectX.DDPixelFormat.dwRGBBitCount == 32)
                        {
                            S9xSetRenderPixelFormat (RGB565);
                            Init_2xSaI (565);
							hq2x_init(16);
                        }
        
        if (
#if 0
            !VOODOO_MODE &&
            !OPENGL_MODE &&
            (
#endif
             (GUI.ScreenDepth == 8 && Settings.SixteenBit) ||
             (GUI.ScreenDepth == 16 && !Settings.SixteenBit) ||
             (GUI.ScreenDepth == 24 || GUI.ScreenDepth == 32))
            GUI.NeedDepthConvert = TRUE;
        
        if (Settings.SixteenBit && 
            (GUI.ScreenDepth == 24 || GUI.ScreenDepth == 32))
        {
            GUI.RedShift += 3;
            GUI.GreenShift += 3;
            GUI.BlueShift += 3;
        }
#if 0
    }
#endif

    int l = 0;
    int i;

    for (i = 0; i < 6; i++)
    {
	int r = (i * 31) / (6 - 1);
	for (int j = 0; j < 6; j++)
	{
	    int g = (j * 31) / (6 - 1);
	    for (int k = 0; k < 6; k++)
	    { 
		int b = (k * 31) / (6 - 1);

		FixedColours [l].red = r << 3;
		FixedColours [l].green = g << 3;
		FixedColours [l++].blue = b << 3;
	    }
	}
    }

    //int *color_diff = new int [0x10000];
    int diffr, diffg, diffb, maxdiff = 0, won = 0, lost;
    int r, d = 8;
    for (r = 0; r <= (int) MAX_RED; r++)
    {
	int cr, g, q;
      
	int k = 6 - 1;
	cr = (r * k) / MAX_RED;
	q  = (r * k) % MAX_RED;
	if (q > d && cr < k) 
	    cr++;
	diffr = abs (cr * k - r);
	for (g = 0; g <= (int) MAX_GREEN; g++)
	{
	    int cg, b;
	  
	    k  = 6 - 1;
	    cg = (g * k) / MAX_GREEN;
	    q  = (g * k) % MAX_GREEN;
	    if(q > d && cg < k)
		cg++;
	    diffg = abs (cg * k - g);
	    for (b = 0; b <= (int) MAX_BLUE; b++) 
	    {
		int cb;
		int rgb = BUILD_PIXEL2(r, g, b);

		k  = 6 - 1;
		cb = (b * k) / MAX_BLUE;
		q  = (b * k) % MAX_BLUE;
		if (q > d && cb < k)
		    cb++;
		diffb = abs (cb * k - b);
		S9xwPalette[rgb] = (cr * 6 + cg) * 6 + cb;
		color_diff[rgb] = diffr + diffg + diffb;
		if (color_diff[rgb] > maxdiff)
		    maxdiff = color_diff[rgb];
	    }
	}
    }

    while (maxdiff > 0 && l < 256)
    {
	int newmaxdiff = 0;
	lost = 0; won++;
	for (r = MAX_RED; r >= 0; r--)
	{
	    int g;
      
	    for (g = MAX_GREEN; g >= 0; g--)
	    {
		int b;
	  
		for (b = MAX_BLUE; b >= 0; b--) 
		{
		    int rgb = BUILD_PIXEL2(r, g, b);

		    if (color_diff[rgb] == maxdiff)
		    {
			if (l >= 256)
			    lost++;
			else
			{
			    FixedColours [l].red = r << 3;
			    FixedColours [l].green = g << 3;
			    FixedColours [l].blue = b << 3;
			    S9xwPalette [rgb] = l++;
			}
			color_diff[rgb] = 0;
		    }
		    else
			if (color_diff[rgb] > newmaxdiff)
			    newmaxdiff = color_diff[rgb];
		    
		}
	    }
	}
	maxdiff = newmaxdiff;
    }
    free (color_diff);
	RenderInit();
}

void S9xAutoSaveSRAM ()
{
    Memory.SaveSRAM (S9xGetFilename (".srm"));
}

void S9xSetPause (uint32 mask)
{
    Settings.ForcedPause |= mask;
}

void S9xClearPause (uint32 mask)
{
    Settings.ForcedPause &= ~mask;
    if (!Settings.ForcedPause)
    {
        // Wake up the main loop thread just if its blocked in a GetMessage
        // call.
        PostMessage (GUI.hWnd, WM_NULL, 0, 0);
    }
}

bool InitUosnesw()
{
	if(!AllocScreenBuffer())
		return FALSE;

    if(!Memory.Init())
		return FALSE;

	if(!DirectX_Initialize(GUI.hWnd, GUI.dwStyleWindowed, GUI.dwStyleFullScreen)) {
		MessageBox(GUI.hWnd,
				   Languages[GUI.Language].errInitDD,
				   "uosnes - DirectX",
				   MB_OK | MB_ICONSTOP);
		return FALSE;
	}
	
	CurFullscreenDisplayModeIndexChange();
	if(!DirectX_SetDisplayMode(DisplayMode[GUI.CurFullscreenDisplayModeIndex].Width,
							   DisplayMode[GUI.CurFullscreenDisplayModeIndex].Height,
							   DisplayMode[GUI.CurFullscreenDisplayModeIndex].Depth,
							   TRUE, GUI.DoubleBuffered)) {
		MessageBox(GUI.hWnd,
				   Languages[GUI.Language].errModeDD,
				   "uosnes - DirectDraw",
				   MB_OK | MB_ICONSTOP);
		DirectX_DeInitialize();
		return FALSE;
	}
	GUI.ScreenDepth = (DWORD)DirectX.Depth;

    S9xSetWinPixelFormat ();

    if(!S9xGraphicsInit())
    	return FALSE;

	S9xInitUpdate();

    if(!S9xInitAPU())
    	return FALSE;

	//init SNES sound
	int _1msc = 0;
	for (; (_1msc * 1000) < Settings.SoundPlaybackRate; _1msc++);
	int buffer_bytes = _1msc * Settings.SoundMixInterval *
		(Settings.Stereo ? 2 : 1) * (Settings.SixteenBitSound ? 2 : 1);
    S9xInitSound(7, Settings.Stereo, buffer_bytes);

	if(!UosneswSetupSound(SoundRates[GUI.SoundMode].rate,
						  SoundRates[GUI.SoundMode].sixteen_bit,
						  SoundRates[GUI.SoundMode].stereo)) {
		MessageBox(GUI.hWnd,
				   Languages[GUI.Language].errInitSound,
				   "uosnes - Sound",
				   MB_OK | MB_ICONINFORMATION);
	}

	return TRUE;
}

void DeinitUosnesw()
{
#ifdef DEBUGGER
    extern FILE *trace;
	extern FILE *trace2;
	if(trace)
		fclose(trace);
	if(trace2)
		fclose(trace2);
#endif
	//Release input resources
	UosneswReleaseInput();
	//Release sound resources
	UosneswReleaseSound();
	//Release DirectX resources
	DirectX_DeInitialize();

	Memory.Deinit();
	S9xGraphicsDeinit();
	S9xDeinitAPU();
	FreeScreenBuffer();
}

void ScreenInit()
{
	if(ScreenBuffer)
		FillMemory ((PVOID)ScreenBuffer, 512*478*2, 0);

	if(SubScreenBuffer)
		FillMemory ((PVOID)SubScreenBuffer, 512*478*2, 0);

	if(ZBuffer)
		FillMemory ((PVOID)ZBuffer, 512*478, 0);

	if(SubZBuffer)
		FillMemory ((PVOID)SubZBuffer, 512*478, 0);

	DirectX_ClearOffscreenSurface();

	//for clear filter log
	ResetRenderMethod();
	RenderInit();
}

static int S9xCompareSDD1IndexEntries (const void *p1, const void *p2)
{
    return (*(uint32 *) p1 - *(uint32 *) p2);
}

void S9xLoadSDD1Data ()
{
    char filename [_MAX_PATH + 1];
    char index [_MAX_PATH + 1];
    char data [_MAX_PATH + 1];

//	Settings.SDD1Pack=FALSE;
    Memory.FreeSDD1Data ();

//    strcpy (filename, GUI.FreezeFileDir);

//	if(strlen(GUI.FreezeFileDir)!=0)
     strcpy (filename, GUI.FreezeFileDir);
/*
	else
	{
	    char dir [_MAX_DIR + 1];
	    char drive [_MAX_DRIVE + 1];
		char name [_MAX_FNAME + 1];
	    char ext [_MAX_EXT + 1];
		mbcs_splitpath(Memory.ROMFilename,drive, dir, name, ext);
		mbcs_makepath(filename,drive, dir, NULL, NULL);
	}
*/

 if(!Settings.SDD1Pack)
 {
    if (strncmp (Memory.ROMName, "Star Ocean", 10) == 0)
	strcat (filename, "\\socnsdd1");
    else
    if (strncmp (Memory.ROMName, "STREET FIGHTER ZERO2", 20) == 0)
	strcat (filename, "\\sfz2sdd1"); // (JAPAN)
    else
    if (strncmp (Memory.ROMName, "STREET FIGHTER ALPHA2", 21) == 0 && Memory.ROM [0x7fd9] == 1)
	strcat (filename, "\\sfa2sdd1"); // (USA)
    else
    if (strncmp (Memory.ROMName, "STREET FIGHTER ALPHA2", 21) == 0 && Memory.ROM [0x7fd9] == 2)
	strcat (filename, "\\sf2esdd1"); // (PAL)
    else
	strcat (filename, "\\unknown");

    strcpy (index, filename);
    strcat (index, "\\SDD1GFX.IDX");
    strcpy (data, filename);
    strcat (data, "\\SDD1GFX.DAT");

    FILE *fs = fopen (index, "rb");
    int len = 0;
    
    if (fs)
    {
        // Index is stored as a sequence of entries, each entry being
        // 12 bytes consisting of:
        // 4 byte key: (24bit address & 0xfffff * 16) | translated block
        // 4 byte ROM offset
        // 4 byte length
        fseek (fs, 0, SEEK_END);
        len = ftell (fs);
        rewind (fs);
        Memory.SDD1Index = (uint8 *) malloc (len);
        fread (Memory.SDD1Index, 1, len, fs);
        fclose (fs);
        Memory.SDD1Entries = len / 12;
        
        if (!(fs = fopen (data, "rb")))
        {
            free ((char *) Memory.SDD1Index);
            Memory.SDD1Index = NULL;
            Memory.SDD1Entries = 0;
        }
        else
        {
            fseek (fs, 0, SEEK_END);
            len = ftell (fs);
            rewind (fs);
            Memory.SDD1Data = (uint8 *) malloc (len);
            fread (Memory.SDD1Data, 1, len, fs);
            fclose (fs);
            
            qsort (Memory.SDD1Index, Memory.SDD1Entries, 12,
                   S9xCompareSDD1IndexEntries);
        }
    }
 }

}

extern "C"
{
	char*osd_GetPackDir()
	{
		static char filename[MAX_PATH];
		memset(filename, 0, MAX_PATH);
		
		if(strlen(GUI.FreezeFileDir)!=0)
			strcpy (filename, GUI.FreezeFileDir);
		else
		{
			char dir [_MAX_DIR + 1];
			char drive [_MAX_DRIVE + 1];
			char name [_MAX_FNAME + 1];
			char ext [_MAX_EXT + 1];
			mbcs_splitpath(Memory.ROMFilename,drive, dir, name, ext);
			mbcs_makepath(filename,drive, dir, NULL, NULL);
		}
		
		if(!strncmp((char*)&Memory.ROM [0xffc0], "SUPER POWER LEAG 4   ", 21))
		{
			strcat(filename, "\\SPL4-SP7");
		}
		else if(!strncmp((char*)&Memory.ROM [0xffc0], "MOMOTETSU HAPPY      ",21))
		{
			strcat(filename, "\\SMHT-SP7");
		}
		else if(!strncmp((char*)&Memory.ROM [0xffc0], "HU TENGAI MAKYO ZERO ", 21))
		{
			strcat(filename, "\\FEOEZSP7");
		}
		else if(!strncmp((char*)&Memory.ROM [0xffc0], "JUMP TENGAIMAKYO ZERO",21))
		{
			strcat(filename, "\\SJUMPSP7");
		}
		else strcat(filename, "\\MISC-SP7");
		return filename;
	}
}

bool JustifierOffscreen()
{
	return (bool)((GUI.MouseButtons&2)!=0);
}

void JustifierButtons(uint32& justifiers)
{
	{
		if((GUI.MouseButtons&1)||(GUI.MouseButtons&2))
		{
			justifiers|=0x00100;
		}
		if(GUI.MouseButtons&4)
		{
			justifiers|=0x00400;
		}
	}
}

static void FrameTimerProc(void *tParams)
{
	if(GUI.PauseEmulation || GUI.StopEmulation || GUI.ForcedPauseEmulation) {
		WaitForSingleObject(GUI.FrameTimerSemaphore, 0);
	}
	else {
		ReleaseSemaphore(GUI.FrameTimerSemaphore, 1, NULL);
	}
	return;
}

BOOL UosneswSetupFrameTimer(DWORD ms)
{
	return SetupFrameTimer(ms,
						   FrameTimerProc,
						   NULL, GUI.hWnd);
}

void DoAVIOpen(const char* filename)
{
}

void DoAVIClose(int reason)
{
}

